from typing import Any


class SafeString:
    """
    A string that will not be escaped when rendered to HTML.
    """

    def __init__(self, safe_str: str) -> None:
        self.safe_str = safe_str

    def __hash__(self) -> int:
        return hash(("SafeString", self.safe_str))

    def __eq__(self, other: Any) -> bool:
        return type(other) is SafeString and other.safe_str == self.safe_str

    def __repr__(self) -> str:
        return f"SafeString('{self.safe_str}')"



import re
from collections.abc import Iterable
from typing import Annotated, Any, Literal, overload

from stario.application import Stario
from stario.exceptions import DatastarConfigError

JSEvent = Literal[
    "abort",
    "afterprint",
    "animationend",
    "animationiteration",
    "animationstart",
    "beforeprint",
    "beforeunload",
    "blur",
    "canplay",
    "canplaythrough",
    "change",
    "click",
    "contextmenu",
    "copy",
    "cut",
    "dblclick",
    "drag",
    "dragend",
    "dragenter",
    "dragleave",
    "dragover",
    "dragstart",
    "drop",
    "durationchange",
    "ended",
    "error",
    "focus",
    "focusin",
    "focusout",
    "fullscreenchange",
    "fullscreenerror",
    "hashchange",
    "input",
    "invalid",
    "keydown",
    "keypress",
    "keyup",
    "load",
    "loadeddata",
    "loadedmetadata",
    "loadstart",
    "message",
    "mousedown",
    "mouseenter",
    "mouseleave",
    "mousemove",
    "mouseover",
    "mouseout",
    "mouseup",
    "mousewheel",
    "offline",
    "online",
    "open",
    "pagehide",
    "pageshow",
    "paste",
    "pause",
    "play",
    "playing",
    "popstate",
    "progress",
    "ratechange",
    "resize",
    "reset",
    "scroll",
    "search",
    "seeked",
    "seeking",
    "select",
    "show",
    "stalled",
    "storage",
    "submit",
    "suspend",
    "timeupdate",
    "toggle",
    "touchcancel",
    "touchend",
    "touchmove",
    "touchstart",
    "transitionend",
    "unload",
    "volumechange",
    "waiting",
    "wheel",
]


# Datastar Actions Literals
ContentType = Literal["json", "form"]
RequestCancellation = Literal["auto", "disabled"]

# Filter types for include/exclude parameters
FilterValue = str | Iterable[str]


type SignalValue = (
    str | int | float | bool | dict[str, SignalValue] | list[SignalValue] | None
)


type TimeValue = int | float | str
"""
We measure time in seconds.
int = number of whole seconds, adds "s" suffix. 10 => 10s.
float = possible fractional seconds, adds "ms" suffix. 0.5 => 500ms.
str = no parsing
"""


def time_to_string(time: TimeValue) -> str:
    """
    Convert a time value to a Datastar-compatible time string.

    Args:
        time: Time value as int (seconds), float (seconds), or string

    Returns:
        Formatted time string for Datastar attributes

    Examples:
        >>> time_to_string(5)
        '5s'
        >>> time_to_string(0.5)
        '500ms'
        >>> time_to_string("2s")
        '2s'
    """
    if isinstance(time, float):
        return f"{int(time * 1000)}ms"
    if isinstance(time, int):
        return f"{int(time)}s"
    return time


type Debounce = (
    TimeValue
    | tuple[TimeValue, Literal["leading", "notrail"]]
    | tuple[TimeValue, Literal["leading", "notrail"], Literal["leading", "notrail"]]
)

type Throttle = (
    TimeValue
    | tuple[TimeValue, Literal["noleading", "trail"]]
    | tuple[TimeValue, Literal["noleading", "trail"], Literal["noleading", "trail"]]
)


def debounce_to_string(debounce: Debounce) -> str:
    """
    Convert a debounce configuration to a Datastar modifier string.

    Debouncing delays the execution of an event handler until after a specified
    time has passed since the last event.

    Args:
        debounce: Time value or tuple with time and modifiers ('leading', 'notrail')

    Returns:
        Datastar debounce modifier string

    Examples:
        >>> debounce_to_string(0.5)
        'debounce.500ms'
        >>> debounce_to_string((1, "leading"))
        'debounce.1s.leading'
        >>> debounce_to_string((0.3, "leading", "notrail"))
        'debounce.300ms.leading.notrail'
    """
    # shortcut for the most common case
    if isinstance(debounce, (int, float, str)):
        return "debounce." + time_to_string(debounce)

    # from typing know it's some tuple:
    if len(debounce) == 2:
        return f"debounce.{time_to_string(debounce[0])}.{debounce[1]}"

    # the last option would be a tuple of 3 elements
    if len(debounce) == 3:
        return f"debounce.{time_to_string(debounce[0])}.{debounce[1]}.{debounce[2]}"

    # or something else that we don't support
    raise DatastarConfigError(
        f"Invalid debounce configuration: {debounce}",
        context={
            "debounce_value": str(debounce),
            "debounce_type": type(debounce).__name__,
        },
        help_text="Debounce must be a time value (int/float/str) or a tuple with time and modifiers.",
        example="""from stario.datastar import Datastar

ds = Datastar()

# Simple debounce (time only):
ds.on("click", "action()", debounce=500)        # 500ms
ds.on("click", "action()", debounce="1s")       # 1 second
ds.on("click", "action()", debounce=1.5)        # 1500ms
# With single modifier:
ds.on("input", "action()", debounce=(300, "notrailing"))

# With two modifiers:
ds.on("input", "action()", debounce=(300, "notrailing", "noleading"))""",
    )


def throttle_to_string(throttle: Throttle) -> str:
    """
    Convert a throttle configuration to a Datastar modifier string.

    Throttling ensures an event handler executes at most once per specified time period.

    Args:
        throttle: Time value or tuple with time and modifiers ('noleading', 'trail')

    Returns:
        Datastar throttle modifier string

    Examples:
        >>> throttle_to_string(1)
        'throttle.1s'
        >>> throttle_to_string((0.5, "trail"))
        'throttle.500ms.trail'
        >>> throttle_to_string((2, "noleading", "trail"))
        'throttle.2s.noleading.trail'
    """
    # shortcut for the most common case
    if isinstance(throttle, (int, float, str)):
        return "throttle." + time_to_string(throttle)

    # from typing know it's some tuple:
    if len(throttle) == 2:
        return f"throttle.{time_to_string(throttle[0])}.{throttle[1]}"

    # the last option would be a tuple of 3 elements
    if len(throttle) == 3:
        return f"throttle.{time_to_string(throttle[0])}.{throttle[1]}.{throttle[2]}"

    # or something else that we don't support
    raise DatastarConfigError(
        f"Invalid throttle configuration: {throttle}",
        context={
            "throttle_value": str(throttle),
            "throttle_type": type(throttle).__name__,
        },
        help_text="Throttle must be a time value (int/float/str) or a tuple with time and modifiers.",
        example="""from stario.datastar import Datastar

ds = Datastar()

# Simple throttle (time only):
ds.on("scroll", "action()", throttle=100)       # 100ms
ds.on("scroll", "action()", throttle="500ms")   # 500ms
ds.on("scroll", "action()", throttle=1.5)       # 1500ms

# With single modifier:
ds.on("scroll", "action()", throttle=(100, "trail"))

# With two modifiers:
ds.on("scroll", "action()", throttle=(100, "noleading", "trail"))""",
    )


def quick_json_dump(obj: dict) -> str:
    """
    Quick JSON dump for small, flat dictionaries used in Datastar attributes.

    This is an optimized JSON serializer for simple key-value pairs commonly
    used in Datastar attributes. Only supports string keys and values.

    Args:
        obj: Dictionary with string keys and values

    Returns:
        JSON string representation

    Examples:
        >>> quick_json_dump({"title": "Hello", "disabled": "true"})
        '{"title":"Hello","disabled":"true"}'
        >>> quick_json_dump({"class": "active"})
        '{"class":"active"}'
    """
    items = [f'"{k}":"{v}"' for k, v in obj.items()]
    return "{" + ",".join(items) + "}"


def parse_filter_value(value: FilterValue) -> str:
    """
    Parse a filter value for include/exclude parameters in Datastar actions.

    Converts filter specifications into regex patterns for signal filtering.
    String values are used as-is (assumed to be regex), while iterables
    are converted to exact-match regex patterns.

    Args:
        value: Either a string (used as-is) or an iterable of strings
                (parsed as regex to match exactly those strings)

    Returns:
        A regex string that matches the filter criteria

    Examples:
        >>> parse_filter_value("foo.*")
        'foo.*'
        >>> parse_filter_value(["foo", "bar"])
        'foo|bar'
        >>> parse_filter_value(("user", "admin"))
        'user|admin'
        >>> parse_filter_value({"count", "total"}) in ('count|total', 'total|count')
        True
        >>> parse_filter_value(iter(["signal1", "signal2"]))
        'signal1|signal2'
    """
    if isinstance(value, str):
        return value

    # Handle any iterable (list, tuple, set, generator, etc.)
    # Escape special regex characters and join with |
    escaped_items = [re.escape(str(item)) for item in value]
    return "|".join(escaped_items)


type Case = Literal["kebab", "snake", "pascal", "camel"]


def to_kebab_key(key: str) -> tuple[str, Case]:
    """
    Convert a key to kebab-case and detect its original casing style.

    Datastar uses kebab-case for attribute names, but supports multiple
    casing styles via the __case modifier. This function converts keys
    to kebab-case and identifies the original case for proper modifier generation.

    Args:
        key: String in any casing style (camelCase, PascalCase, snake_case, kebab-case)

    Returns:
        Tuple of (kebab-case string, original case style)

    Examples:
        >>> to_kebab_key("mySignal")
        ('my-signal', 'camel')
        >>> to_kebab_key("MySignal")
        ('my-signal', 'pascal')
        >>> to_kebab_key("my_signal")
        ('my-signal', 'snake')
        >>> to_kebab_key("my-signal")
        ('my-signal', 'kebab')
        >>> to_kebab_key("signal")
        ('signal', 'kebab')
    """
    # snake_case (contains '_') - check before lowercase check
    if "_" in key:
        return key.replace("_", "-").lower(), "snake"

    # kebab-case (contains '-')
    if "-" in key:
        return key.lower(), "kebab"

    # all lowercase (already kebab)
    if key.islower():
        return key, "kebab"

    # PascalCase (first char uppercase)
    # Insert dash before every uppercase except first, then lowercase
    if key[0].isupper():
        return (
            "".join(
                (
                    ("-" if i != 0 and c.isupper() else "") + c.lower()
                    for i, c in enumerate(key)
                )
            ),
            "pascal",
        )

    # camelCase (first char lowercase, has uppercase somewhere)
    # Insert dash before every uppercase, then lowercase
    return (
        "".join((("-" + c.lower()) if c.isupper() else c for c in key)),
        "camel",
    )


class DatastarAttributes:
    """
    Generator for Datastar data-* attributes and actions.

    This class provides methods to generate all Datastar attributes and actions
    with proper formatting, modifiers, and type safety. It integrates with the
    Stario application to resolve endpoint names to URLs.

    Reference: https://data-star.dev/reference/attributes

    Attributes:
        app: The Stario application instance for URL resolution

    Examples:
        >>> from stario.application import Stario
        >>> app = Stario()
        >>> ds = DatastarAttributes(app)
        >>> ds.text("$message")
        {'data-text': '$message'}
        >>> ds.show("$isVisible")
        {'data-show': '$isVisible'}
    """

    def __init__(self, app: Stario) -> None:
        """
        Initialize the DatastarAttributes generator.

        Args:
            app: Stario application instance for endpoint name -> URL resolution
        """
        self.app = app

    def attr(self, attr_dict: dict[str, str]) -> dict[str, str]:
        """
        Set the value of HTML attributes to expressions and keep them in sync.

        The data-attr attribute dynamically updates HTML attribute values based on
        signal expressions. When signals change, the attributes automatically update.

        Reference: https://data-star.dev/reference/attributes#data-attr

        Args:
            attr_dict: Dictionary mapping attribute names to Datastar expressions

        Returns:
            Dictionary with data-attr attribute

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.attr({"title": "$tooltip", "disabled": "$isDisabled"})
            {'data-attr': '{"title":"$tooltip","disabled":"$isDisabled"}'}
        """
        return {"data-attr": quick_json_dump(attr_dict)}

    def bind(self, signal_name: str) -> dict[str, str]:
        """
        Set up two-way data binding between a signal and an element's value.

        Creates a signal (if it doesn't exist) and synchronizes it with the element's
        value. Updates flow both ways: signal changes update the element, and element
        changes (via user input) update the signal.

        Reference: https://data-star.dev/reference/attributes#data-bind

        Args:
            signal_name: Name of the signal to bind to

        Returns:
            Dictionary with data-bind attribute

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.bind("username")
            {'data-bind': 'username'}
            >>> ds.bind("email")
            {'data-bind': 'email'}
        """
        return {"data-bind": signal_name}

    def class_(self, class_dict: dict[str, str]) -> dict[str, str]:
        """
        Add or remove CSS classes based on expressions.

        Each key-value pair represents a class name and a boolean expression.
        When the expression evaluates to true, the class is added; otherwise removed.

        Reference: https://data-star.dev/reference/attributes#data-class

        Args:
            class_dict: Dictionary mapping class names to boolean expressions

        Returns:
            Dictionary with data-class attribute

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.class_({"hidden": "$isHidden", "active": "$isActive"})
            {'data-class': '{"hidden":"$isHidden","active":"$isActive"}'}
        """
        return {"data-class": quick_json_dump(class_dict)}

    def computed(self, computed_dict: dict[str, str]) -> dict[str, str]:
        """
        Create read-only computed signals based on expressions.

        Computed signals automatically recalculate when their dependencies change.
        They're useful for memoizing derived values and avoiding redundant calculations.
        Signal names are automatically converted to camelCase.

        Reference: https://data-star.dev/reference/attributes#data-computed

        Args:
            computed_dict: Dictionary mapping signal names to expressions

        Returns:
            Dictionary with data-computed-* attributes with proper case modifiers

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.computed({"fullName": "$firstName + ' ' + $lastName"})
            {'data-computed-full-name': "$firstName + ' ' + $lastName"}
            >>> ds.computed({"total_price": "$quantity * $price"})
            {'data-computed-total-price__case.snake': '$quantity * $price'}
        """

        # 2-pass: first collect kebab/from_case, then build dict
        kebab_cases = [
            (to_kebab_key(key), value) for key, value in computed_dict.items()
        ]
        result = {
            (
                f"data-computed-{kebab_key}"
                if from_case == "camel"  # Default
                else f"data-computed-{kebab_key}__case.{from_case}"
            ): value
            for (kebab_key, from_case), value in kebab_cases
        }

        return result

    def effect(self, expression: str) -> dict[str, str]:
        """
        Execute an expression on load and whenever dependencies change.

        Effects run immediately on page load and re-run whenever any signals
        referenced in the expression change. Use for side effects like updating
        other signals, making backend requests, or DOM manipulation.

        Reference: https://data-star.dev/reference/attributes#data-effect

        Args:
            expression: Datastar expression to execute

        Returns:
            Dictionary with data-effect attribute

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.effect("$total = $price * $quantity")
            {'data-effect': '$total = $price * $quantity'}
            >>> ds.effect("console.log($count)")
            {'data-effect': 'console.log($count)'}
        """
        return {"data-effect": expression}

    def ignore(self, self_only: bool = False) -> dict[str, bool]:
        """
        Tell Datastar to ignore an element and optionally its descendants.

        Prevents Datastar from processing data-* attributes on the element.
        Useful for avoiding conflicts with third-party libraries or when
        displaying user-generated content.

        Reference: https://data-star.dev/reference/attributes#data-ignore

        Args:
            self_only: If True, only ignore the element itself, not its children

        Returns:
            Dictionary with data-ignore or data-ignore__self attribute

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.ignore()
            {'data-ignore': True}
            >>> ds.ignore(self_only=True)
            {'data-ignore__self': True}
        """
        if self_only:
            return {"data-ignore__self": True}
        return {"data-ignore": True}

    def ignore_morph(self) -> dict[str, bool]:
        """
        Tell Datastar to skip morphing this element during DOM updates.

        Prevents the element and its children from being updated when the
        PatchElements watcher processes DOM morphing operations.

        Reference: https://data-star.dev/reference/attributes#data-ignore-morph

        Returns:
            Dictionary with data-ignore-morph attribute

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.ignore_morph()
            {'data-ignore-morph': True}
        """
        return {"data-ignore-morph": True}

    def indicator(self, signal_name: str) -> dict[str, str]:
        """
        Create a boolean signal that tracks fetch request status.

        The signal is automatically set to true when a fetch request starts
        and false when it completes. Useful for showing loading indicators.

        Reference: https://data-star.dev/reference/attributes#data-indicator

        Args:
            signal_name: Name of the indicator signal to create

        Returns:
            Dictionary with data-indicator attribute

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.indicator("loading")
            {'data-indicator': 'loading'}
            >>> ds.indicator("isFetching")
            {'data-indicator': 'isFetching'}
        """
        return {"data-indicator": signal_name}

    def json_signals(
        self,
        *,
        include: FilterValue | None = None,
        exclude: FilterValue | None = None,
        terse: bool = False,
    ) -> dict[str, str | bool]:
        """
        Create an element containing JSON representation of signals.

        Serializes signals to JSON within the element's content. Useful for
        passing signal state to JavaScript or debugging. Can filter signals
        using regex patterns.

        Reference: https://data-star.dev/reference/attributes#data-json-signals

        Args:
            include: Regex pattern or collection to match signal names to include
            exclude: Regex pattern or collection to match signal names to exclude
            terse: If True, minify the JSON output

        Returns:
            Dictionary with data-json-signals attribute

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.json_signals()
            {'data-json-signals': True}
            >>> ds.json_signals(include="user.*")
            {'data-json-signals': '{"include":"user.*"}'}
            >>> ds.json_signals(exclude=["password", "token"], terse=True)
            {'data-json-signals__terse': '{"exclude":"password|token"}'}
        """

        # Fastest possible: minimize branches, dict creation, and function calls.
        # Use local vars, avoid repeated lookups, and combine logic.
        if include is not None or exclude is not None:
            d = {}
            if include is not None:
                d["include"] = parse_filter_value(include)
            if exclude is not None:
                d["exclude"] = parse_filter_value(exclude)
            value = quick_json_dump(d)
        else:
            value = True

        if terse:
            return {"data-json-signals__terse": value}
        return {"data-json-signals": value}

    @overload
    def on(
        self,
        event: Literal["intersect"],
        expression: str,
        *,
        once: bool = False,
        half: bool = False,
        full: bool = False,
        delay: TimeValue | None = None,
        debounce: Debounce | None = None,
        throttle: Throttle | None = None,
        viewtransition: bool = False,
    ) -> dict[str, str]:
        """
        Execute an expression when an element intersects with the viewport.

        Triggers when the element becomes visible (intersects with viewport).
        Uses the Intersection Observer API under the hood.

        Reference: https://data-star.dev/reference/attributes#data-on-intersect

        Args:
            expression: Datastar expression to execute
            once: Fire only once, then remove the listener
            half: Trigger when at least 50% is visible
            full: Trigger when 100% is visible
            delay: Delay before executing expression
            debounce: Debounce configuration
            throttle: Throttle configuration
            viewtransition: Use view transitions for updates

        Returns:
            Dictionary with data-on-intersect attribute

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.on("intersect", "console.log('visible')")
            {'data-on-intersect': "console.log('visible')"}
            >>> ds.on("intersect", "$load()", once=True, half=True)
            {'data-on-intersect__once__half': '$load()'}
        """

    @overload
    def on(
        self,
        event: Literal["interval"],
        expression: str,
        *,
        duration: TimeValue | tuple[TimeValue, Literal["leading"]] = "1s",
        viewtransition: bool = False,
    ) -> dict[str, str]:
        """
        Execute an expression at regular intervals.

        Runs the expression repeatedly at the specified interval. By default,
        waits for the first interval before executing (trailing edge). Use
        'leading' modifier to execute immediately on page load.

        Reference: https://data-star.dev/reference/attributes#data-on-interval

        Args:
            expression: Datastar expression to execute
            duration: Interval duration or tuple with duration and "leading" modifier
            viewtransition: Use view transitions for updates

        Returns:
            Dictionary with data-on-interval attribute

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.on("interval", "$refresh()")
            {'data-on-interval': '$refresh()'}
            >>> ds.on("interval", "$poll()", duration=5)
            {'data-on-interval__duration.5s': '$poll()'}
            >>> ds.on("interval", "$check()", duration=(2, "leading"))
            {'data-on-interval__duration.2s.leading': '$check()'}
        """

    @overload
    def on(
        self,
        event: Literal["load"],
        expression: str,
        *,
        delay: TimeValue | None = None,
        viewtransition: bool = False,
    ) -> dict[str, str]:
        """
        Execute an expression when the element loads.

        Triggers once when the element is first added to the DOM and
        Datastar initializes it. Useful for lazy-loading data.

        Reference: https://data-star.dev/reference/attributes#data-on-load

        Args:
            expression: Datastar expression to execute
            delay: Delay before executing expression
            viewtransition: Use view transitions for updates

        Returns:
            Dictionary with data-on-load attribute

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.on("load", "$init()")
            {'data-on-load': '$init()'}
            >>> ds.on("load", "$fetchData()", delay=1)
            {'data-on-load__delay.1s': '$fetchData()'}
        """

    @overload
    def on(
        self,
        event: Literal["signal-patch"],
        expression: str,
        *,
        delay: TimeValue | None = None,
        debounce: Debounce | None = None,
        throttle: Throttle | None = None,
        # Filter
        include: FilterValue | None = None,
        exclude: FilterValue | None = None,
    ) -> dict[str, str]:
        """
        Execute an expression when signals are patched from the backend.

        Triggers when server-sent events (SSE) update signals. Can be filtered
        to only react to specific signal changes.

        Reference: https://data-star.dev/reference/attributes#data-on-signal-patch

        Args:
            expression: Datastar expression to execute
            delay: Delay before executing expression
            debounce: Debounce configuration
            throttle: Throttle configuration
            include: Only trigger for signals matching this pattern
            exclude: Don't trigger for signals matching this pattern

        Returns:
            Dictionary with data-on-signal-patch attribute(s)

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.on("signal-patch", "console.log('updated')")
            {'data-on-signal-patch': "console.log('updated')"}
            >>> ds.on("signal-patch", "$log()", include="user.*", debounce=0.5)
            {'data-on-signal-patch__debounce.500ms': '$log()', 'data-on-signal-patch-filter': '{"include":"user.*"}'}
        """

    @overload
    def on(
        self,
        event: JSEvent | str,
        expression: str,
        *,
        once: bool = False,
        passive: bool = False,
        capture: bool = False,
        delay: TimeValue | None = None,
        debounce: Debounce | None = None,
        throttle: Throttle | None = None,
        viewtransition: bool = False,
        window: bool = False,
        outside: bool = False,
        prevent: bool = False,
        stop: bool = False,
    ) -> dict[str, str]:
        """
        Execute an expression in response to a DOM event.

        Attaches an event listener to the element and executes the expression
        when the event fires. Supports all standard DOM events and custom events.

        Reference: https://data-star.dev/reference/attributes#data-on

        Args:
            event: Event name (click, input, submit, etc.) or custom event name
            expression: Datastar expression to execute
            once: Fire only once, then remove the listener
            passive: Mark listener as passive (improves scroll performance)
            capture: Use capture phase instead of bubble phase
            delay: Delay before executing expression
            debounce: Debounce configuration
            throttle: Throttle configuration
            viewtransition: Use view transitions for updates
            window: Attach listener to window instead of element
            outside: Trigger when event occurs outside the element
            prevent: Call preventDefault() on the event
            stop: Call stopPropagation() on the event

        Returns:
            Dictionary with data-on-{event} attribute

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.on("click", "$count++")
            {'data-on-click': '$count++'}
            >>> ds.on("submit", "$save()", prevent=True)
            {'data-on-submit__prevent': '$save()'}
            >>> ds.on("input", "$search()", debounce=0.3)
            {'data-on-input__debounce.300ms': '$search()'}
            >>> ds.on("click", "$close()", outside=True)
            {'data-on-click__outside': '$close()'}
            >>> ds.on("scroll", "$track()", window=True, throttle=0.1)
            {'data-on-scroll__window__throttle.100ms': '$track()'}
        """

    def on(
        self,
        event: str,
        expression: str,
        **kwargs: Any,
    ) -> dict[str, str]:

        if event == "intersect":
            return self._on_intersect(expression, **kwargs)
        if event == "interval":
            return self._on_interval(expression, **kwargs)
        if event == "load":
            return self._on_load(expression, **kwargs)
        if event == "signal-patch":
            return self._on_signal_patch(expression, **kwargs)
        return self._on_event(event, expression, **kwargs)

    def _on_intersect(
        self,
        expression: str,
        once: bool = False,
        half: bool = False,
        full: bool = False,
        delay: TimeValue | None = None,
        debounce: Debounce | None = None,
        throttle: Throttle | None = None,
        viewtransition: bool = False,
    ) -> dict[str, str]:

        modifiers: list[str] = []
        append = modifiers.append
        if once:
            append("once")
        if half:
            append("half")
        if full:
            append("full")
        if delay is not None:
            append("delay." + time_to_string(delay))
        if debounce is not None:
            append(debounce_to_string(debounce))
        if throttle is not None:
            append(throttle_to_string(throttle))
        if viewtransition:
            append("viewtransition")

        if len(modifiers) > 0:
            mods = "__".join(modifiers)
            return {"data-on-intersect__" + mods: expression}

        return {"data-on-intersect": expression}

    def _on_interval(
        self,
        expression: str,
        duration: TimeValue | tuple[TimeValue, Literal["leading"]] = "1s",
        viewtransition: bool = False,
    ) -> dict[str, str]:

        if duration == "1s":
            if not viewtransition:
                return {"data-on-interval": expression}  # default

            return {"data-on-interval__viewtransition": expression}

        # TimeValue is a type alias (int | float | str), so we check for those types explicitly:
        if isinstance(duration, (int, float, str)):
            mods = "duration." + time_to_string(duration)

        elif isinstance(duration, tuple):
            mods = f"duration.{time_to_string(duration[0])}.{duration[1]}"

        else:
            raise DatastarConfigError(
                f"Invalid duration configuration for on_interval: {duration}",
                context={
                    "duration_value": str(duration),
                    "duration_type": type(duration).__name__,
                },
                help_text="Duration must be a time value (int/float/str) or a tuple with time and 'leading' modifier.",
                example="""from stario.datastar import Datastar

ds = Datastar()

# Simple duration (time only):
ds.on_interval("action()", duration=1000)      # 1000ms (1 second)
ds.on_interval("action()", duration="5s")      # 5 seconds
ds.on_interval("action()", duration=2.5)       # 2.5 seconds

# With leading modifier (fires immediately, then at intervals):
ds.on_interval("action()", duration=(1000, "leading"))""",
            )

        if viewtransition:
            mods += "__viewtransition"

        return {"data-on-interval__" + mods: expression}

    def _on_load(
        self,
        expression: str,
        delay: TimeValue | None = None,
        viewtransition: bool = False,
    ) -> dict[str, str]:

        if delay is None:
            if not viewtransition:
                return {"data-on-load": expression}  # default

            return {"data-on-load__viewtransition": expression}

        mods = "delay." + time_to_string(delay)

        if viewtransition:
            mods += "__viewtransition"

        return {"data-on-load__" + mods: expression}

    def _on_signal_patch(
        self,
        expression: str,
        delay: TimeValue | None = None,
        debounce: Debounce | None = None,
        throttle: Throttle | None = None,
        # Filter
        include: FilterValue | None = None,
        exclude: FilterValue | None = None,
    ) -> dict[str, str]:

        modifiers: list[str] = []
        append = modifiers.append

        if delay is not None:
            append("delay." + time_to_string(delay))

        if debounce is not None:
            append(debounce_to_string(debounce))

        if throttle is not None:
            append(throttle_to_string(throttle))

        if len(modifiers) > 0:
            key = "data-on-signal-patch__" + "__".join(modifiers)
        else:
            key = "data-on-signal-patch"

        if include is not None or exclude is not None:
            filter_dict = {}
            if include is not None:
                filter_dict["include"] = include
            if exclude is not None:
                filter_dict["exclude"] = exclude

            return {
                key: expression,
                "data-on-signal-patch-filter": quick_json_dump(filter_dict),
            }

        return {key: expression}

    def _on_event(
        self,
        event: JSEvent | str,
        expression: str,
        *,
        once: bool = False,
        passive: bool = False,
        capture: bool = False,
        delay: TimeValue | None = None,
        debounce: Debounce | None = None,
        throttle: Throttle | None = None,
        viewtransition: bool = False,
        window: bool = False,
        outside: bool = False,
        prevent: bool = False,
        stop: bool = False,
    ) -> dict[str, str]:

        modifiers = []
        append = modifiers.append
        if once:
            append("once")
        if passive:
            append("passive")
        if capture:
            append("capture")
        if window:
            append("window")
        if outside:
            append("outside")
        if prevent:
            append("prevent")
        if stop:
            append("stop")
        if delay is not None:
            append("delay." + time_to_string(delay))
        if debounce is not None:
            append(debounce_to_string(debounce))
        if throttle is not None:
            append(throttle_to_string(throttle))
        if viewtransition:
            append("viewtransition")

        kebab_event, from_case = to_kebab_key(event)
        if from_case != "kebab":  # Default
            append("case." + from_case)

        if len(modifiers) > 0:
            mods = "__".join(modifiers)
            return {f"data-on-{kebab_event}__{mods}": expression}

        return {f"data-on-{kebab_event}": expression}

    def preserve_attr(self, attrs: str | list[str]) -> dict[str, str]:
        """
        Preserve client-side attribute values during DOM morphing.

        Prevents specified attributes from being overwritten when elements
        are morphed/updated from backend responses. Useful for maintaining
        client-side state like focus or scroll position.

        Reference: https://data-star.dev/reference/attributes#data-preserve-attr

        Args:
            attrs: Attribute name(s) to preserve

        Returns:
            Dictionary with data-preserve-attr attribute

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.preserve_attr("value")
            {'data-preserve-attr': 'value'}
            >>> ds.preserve_attr(["value", "checked", "disabled"])
            {'data-preserve-attr': 'value checked disabled'}
        """
        value = attrs if isinstance(attrs, str) else " ".join(attrs)
        return {"data-preserve-attr": value}

    def ref(self, signal_name: str) -> dict[str, str]:
        """
        Create a signal that references the DOM element.

        Stores a reference to the element in a signal, allowing access to
        the DOM node from Datastar expressions and effects.

        Reference: https://data-star.dev/reference/attributes#data-ref

        Args:
            signal_name: Name of the signal to store the element reference

        Returns:
            Dictionary with data-ref attribute

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.ref("inputEl")
            {'data-ref': 'inputEl'}
            >>> ds.ref("modalElement")
            {'data-ref': 'modalElement'}
        """
        return {"data-ref": signal_name}

    def show(self, expression: str) -> dict[str, str]:
        """
        Show or hide an element based on a boolean expression.

        When the expression evaluates to true, the element is visible;
        when false, it's hidden (display: none). Reactively updates when
        signals change.

        Reference: https://data-star.dev/reference/attributes#data-show

        Args:
            expression: Boolean Datastar expression

        Returns:
            Dictionary with data-show attribute

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.show("$isVisible")
            {'data-show': '$isVisible'}
            >>> ds.show("$count > 0")
            {'data-show': '$count > 0'}
        """
        return {"data-show": expression}

    def signals(
        self,
        signals_dict: dict[str, SignalValue],
        *,
        ifmissing: bool = False,
    ) -> dict[str, str]:
        """
        Initialize signals with values.

        Creates or updates signals with specified initial values. Signals
        are reactive data stores that trigger updates when changed.

        Reference: https://data-star.dev/reference/attributes#data-signals

        Args:
            signals_dict: Dictionary mapping signal names to initial values
            ifmissing: If True, only set values for signals that don't exist

        Returns:
            Dictionary with data-signals or data-signals__ifmissing attribute

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.signals({"count": "0", "name": "John"})
            {'data-signals': '{"count":"0","name":"John"}'}
            >>> ds.signals({"fallback": "default"}, ifmissing=True)
            {'data-signals__ifmissing': '{"fallback":"default"}'}
        """
        if ifmissing:
            return {"data-signals__ifmissing": quick_json_dump(signals_dict)}
        return {"data-signals": quick_json_dump(signals_dict)}

    def style(self, style_dict: dict[str, str]) -> dict[str, str]:
        """
        Set CSS styles based on expressions.

        Each key-value pair represents a CSS property and an expression for
        its value. Styles update reactively when signals change.

        Reference: https://data-star.dev/reference/attributes#data-style

        Args:
            style_dict: Dictionary mapping CSS properties to expressions

        Returns:
            Dictionary with data-style attribute

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.style({"color": "$themeColor", "opacity": "$alpha"})
            {'data-style': '{"color":"$themeColor","opacity":"$alpha"}'}
        """
        return {"data-style": quick_json_dump(style_dict)}

    def text(self, expression: str) -> dict[str, str]:
        """
        Set the text content of an element from an expression.

        Updates the element's text content reactively when the expression
        changes. Automatically escapes HTML to prevent XSS.

        Reference: https://data-star.dev/reference/attributes#data-text

        Args:
            expression: Datastar expression evaluating to a string

        Returns:
            Dictionary with data-text attribute

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.text("$message")
            {'data-text': '$message'}
            >>> ds.text("$firstName + ' ' + $lastName")
            {'data-text': "$firstName + ' ' + $lastName"}
        """
        return {"data-text": expression}

    # Datastar Actions
    def peek(self, callable_expr: str) -> str:
        """
        Access signals without subscribing to their changes.

        Reads signal values without triggering re-evaluation when those signals
        change. Useful for one-time reads or breaking circular dependencies.

        Reference: https://data-star.dev/reference/actions#peek

        Args:
            callable_expr: Expression to evaluate without creating subscriptions

        Returns:
            Datastar @peek() action string

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.peek("$count")
            '@peek($count)'
            >>> ds.peek("$user.name")
            '@peek($user.name)'
        """
        return f"@peek({callable_expr})"

    def set_all(
        self,
        value: str,
        include: FilterValue | None = None,
        exclude: FilterValue | None = None,
    ) -> str:
        """
        Set the value of multiple signals at once.

        Updates all matching signals (or all signals if no filter specified)
        to the same value. Useful for bulk resets or initializations.

        Reference: https://data-star.dev/reference/actions#setall

        Args:
            value: Expression for the new value
            include: Regex pattern or collection to match signal names to include
            exclude: Regex pattern or collection to match signal names to exclude

        Returns:
            Datastar @setAll() action string

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.set_all("0")
            '@setAll(0)'
            >>> ds.set_all("false", include="is.*")
            '@setAll(false, {"include":"is.*"})'
            >>> ds.set_all("''", exclude=["id", "token"])
            '@setAll(\\'\\', {"exclude":"id|token"})'
        """
        if include is not None or exclude is not None:
            filter_dict = {}
            if include is not None:
                filter_dict["include"] = parse_filter_value(include)
            if exclude is not None:
                filter_dict["exclude"] = parse_filter_value(exclude)
            filter_json = quick_json_dump(filter_dict)

            return f"@setAll({value}, {filter_json})"
        return f"@setAll({value})"

    def toggle_all(
        self, include: FilterValue | None = None, exclude: FilterValue | None = None
    ) -> str:
        """
        Toggle the boolean value of multiple signals.

        Flips boolean values of all matching signals (or all signals if no
        filter specified). True becomes false, false becomes true.

        Reference: https://data-star.dev/reference/actions#toggleall

        Args:
            include: Regex pattern or collection to match signal names to include
            exclude: Regex pattern or collection to match signal names to exclude

        Returns:
            Datastar @toggleAll() action string

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.toggle_all()
            '@toggleAll()'
            >>> ds.toggle_all(include="is.*")
            '@toggleAll({"include":"is.*"})'
            >>> ds.toggle_all(exclude=["isAdmin", "isRoot"])
            '@toggleAll({"exclude":"isAdmin|isRoot"})'
        """
        if include is not None or exclude is not None:
            filter_dict = {}
            if include is not None:
                filter_dict["include"] = parse_filter_value(include)
            if exclude is not None:
                filter_dict["exclude"] = parse_filter_value(exclude)
            filter_json = quick_json_dump(filter_dict)

            return f"@toggleAll({filter_json})"
        return "@toggleAll()"

    def _http_action(
        self,
        method: str,
        uri: str,
        *,
        content_type: ContentType | str = "json",
        include: FilterValue | None = None,
        exclude: FilterValue | None = None,
        selector: str | None = None,
        headers: dict[str, str] | None = None,
        open_when_hidden: bool = False,
        retry_interval_ms: int = 1_000,
        retry_scaler: float = 2.0,
        retry_max_wait_ms: int = 30000,
        retry_max_count: int = 10,
        request_cancellation: RequestCancellation | str = "auto",
    ) -> str:
        """
        Internal: Generate HTTP action strings for fetch requests.

        Base method used by get(), post(), put(), patch(), and delete() to
        generate properly formatted Datastar HTTP action strings with options.

        Args:
            method: HTTP method (get, post, put, patch, delete)
            uri: Endpoint URL or name (resolved via app.url_path_for if no leading /)
            content_type: Request content type (json or form)
            include: Signal filter for inclusion in request
            exclude: Signal filter for exclusion from request
            selector: CSS selector for target elements to update
            headers: Additional HTTP headers
            open_when_hidden: Open SSE connection even when page is hidden
            retry_interval_ms: Initial retry interval in milliseconds
            retry_scaler: Multiplier for exponential backoff
            retry_max_wait_ms: Maximum retry interval
            retry_max_count: Maximum number of retries
            request_cancellation: Request cancellation strategy

        Returns:
            Formatted Datastar HTTP action string
        """
        options = []

        # Only add content_type if it's not the default
        if content_type != "json":
            options.append(f"contentType: '{content_type}'")

        # Handle include/exclude filters
        if include is not None or exclude is not None:
            filter_dict = {}
            if include is not None:
                filter_dict["include"] = parse_filter_value(include)
            if exclude is not None:
                filter_dict["exclude"] = parse_filter_value(exclude)

            filter_json = quick_json_dump(filter_dict)
            options.append(f"filterSignals: {filter_json}")

        if selector is not None:
            options.append(f"selector: '{selector}'")

        if headers is not None:
            headers_json = quick_json_dump(headers)
            options.append(f"headers: {headers_json}")

        if open_when_hidden:
            options.append("openWhenHidden: true")

        # Only add retry options if they're not the defaults
        if retry_interval_ms != 1_000:
            options.append(f"retryInterval: {retry_interval_ms}")

        if retry_scaler != 2.0:
            options.append(f"retryScaler: {retry_scaler}")

        if retry_max_wait_ms != 30_000:
            options.append(f"retryMaxWaitMs: {retry_max_wait_ms}")

        if retry_max_count != 10:
            options.append(f"retryMaxCount: {retry_max_count}")

        if request_cancellation != "auto":
            options.append(f"requestCancellation: '{request_cancellation}'")

        if uri[0] != "/":
            # without leading slash, we assume this is the name of the endpoint
            uri = self.app.url_path_for(uri)

        if options:
            options_str = "{" + ", ".join(options) + "}"
            return f"@{method}('{uri}', {options_str})"
        else:
            return f"@{method}('{uri}')"

    def get(
        self,
        uri: str,
        *,
        content_type: ContentType | str = "json",
        include: FilterValue | None = None,
        exclude: FilterValue | None = None,
        selector: str | None = None,
        headers: dict[str, str] | None = None,
        open_when_hidden: bool = False,
        retry_interval_ms: int = 1_000,
        retry_scaler: float = 2.0,
        retry_max_wait_ms: int = 30_000,
        retry_max_count: int = 10,
        request_cancellation: RequestCancellation | str = "auto",
    ) -> str:
        """
        Send a GET request to fetch data from the backend.

        Initiates a GET request and processes the SSE response stream to
        update signals and morph DOM elements.

        Reference: https://data-star.dev/reference/actions#get

        Args:
            uri: Endpoint URL or route name
            content_type: Request content type
            include: Include only matching signals in request
            exclude: Exclude matching signals from request
            selector: CSS selector for elements to update
            headers: Additional HTTP headers
            open_when_hidden: Keep connection open when page is hidden
            retry_interval_ms: Initial retry delay
            retry_scaler: Exponential backoff multiplier
            retry_max_wait_ms: Maximum retry delay
            retry_max_count: Maximum retry attempts
            request_cancellation: Cancellation strategy

        Returns:
            Datastar @get() action string

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.get("/api/users")
            "@get('/api/users')"
            >>> ds.get("/api/data", include="user.*")
            '@get(\\'/api/data\\', {filterSignals: {"include":"user.*"}})'
        """
        return self._http_action(
            "get",
            uri,
            content_type=content_type,
            include=include,
            exclude=exclude,
            selector=selector,
            headers=headers,
            open_when_hidden=open_when_hidden,
            retry_interval_ms=retry_interval_ms,
            retry_scaler=retry_scaler,
            retry_max_wait_ms=retry_max_wait_ms,
            retry_max_count=retry_max_count,
            request_cancellation=request_cancellation,
        )

    def post(
        self,
        uri: str,
        *,
        content_type: ContentType | str = "json",
        include: FilterValue | None = None,
        exclude: FilterValue | None = None,
        selector: str | None = None,
        headers: dict[str, str] | None = None,
        open_when_hidden: bool = False,
        retry_interval_ms: int = 1_000,
        retry_scaler: float = 2.0,
        retry_max_wait_ms: int = 30_000,
        retry_max_count: int = 10,
        request_cancellation: RequestCancellation | str = "auto",
    ) -> str:
        """
        Send a POST request with signal data to the backend.

        Initiates a POST request with signals as payload, processes SSE response
        stream to update signals and morph DOM elements.

        Reference: https://data-star.dev/reference/actions#post

        Args:
            uri: Endpoint URL or route name
            content_type: Request content type ('json' or 'form')
            include: Include only matching signals in request
            exclude: Exclude matching signals from request
            selector: CSS selector for elements to update
            headers: Additional HTTP headers
            open_when_hidden: Keep connection open when page is hidden
            retry_interval_ms: Initial retry delay
            retry_scaler: Exponential backoff multiplier
            retry_max_wait_ms: Maximum retry delay
            retry_max_count: Maximum retry attempts
            request_cancellation: Cancellation strategy

        Returns:
            Datastar @post() action string

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.post("/api/users")
            "@post('/api/users')"
            >>> ds.post("/api/form", content_type="form")
            "@post('/api/form', {contentType: 'form'})"
        """
        return self._http_action(
            "post",
            uri,
            content_type=content_type,
            include=include,
            exclude=exclude,
            selector=selector,
            headers=headers,
            open_when_hidden=open_when_hidden,
            retry_interval_ms=retry_interval_ms,
            retry_scaler=retry_scaler,
            retry_max_wait_ms=retry_max_wait_ms,
            retry_max_count=retry_max_count,
            request_cancellation=request_cancellation,
        )

    def put(
        self,
        uri: str,
        *,
        content_type: ContentType | str = "json",
        include: FilterValue | None = None,
        exclude: FilterValue | None = None,
        selector: str | None = None,
        headers: dict[str, str] | None = None,
        open_when_hidden: bool = False,
        retry_interval_ms: int = 1_000,
        retry_scaler: float = 2.0,
        retry_max_wait_ms: int = 30_000,
        retry_max_count: int = 10,
        request_cancellation: RequestCancellation | str = "auto",
    ) -> str:
        """
        Send a PUT request to update resources on the backend.

        Reference: https://data-star.dev/reference/actions#put

        Args:
            uri: Endpoint URL or route name
            (See post() for full parameter documentation)

        Returns:
            Datastar @put() action string

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.put("/api/users/123")
            "@put('/api/users/123')"
        """
        return self._http_action(
            "put",
            uri,
            content_type=content_type,
            include=include,
            exclude=exclude,
            selector=selector,
            headers=headers,
            open_when_hidden=open_when_hidden,
            retry_interval_ms=retry_interval_ms,
            retry_scaler=retry_scaler,
            retry_max_wait_ms=retry_max_wait_ms,
            retry_max_count=retry_max_count,
            request_cancellation=request_cancellation,
        )

    def patch(
        self,
        uri: str,
        *,
        content_type: ContentType | str = "json",
        include: FilterValue | None = None,
        exclude: FilterValue | None = None,
        selector: str | None = None,
        headers: dict[str, str] | None = None,
        open_when_hidden: bool = False,
        retry_interval_ms: int = 1_000,
        retry_scaler: float = 2.0,
        retry_max_wait_ms: int = 30_000,
        retry_max_count: int = 10,
        request_cancellation: RequestCancellation | str = "auto",
    ) -> str:
        """
        Send a PATCH request to partially update resources on the backend.

        Reference: https://data-star.dev/reference/actions#patch

        Args:
            uri: Endpoint URL or route name
            (See post() for full parameter documentation)

        Returns:
            Datastar @patch() action string

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.patch("/api/users/123", include=["name", "email"])
            '@patch(\\'/api/users/123\\', {filterSignals: {"include":"name|email"}})'
        """
        return self._http_action(
            "patch",
            uri,
            content_type=content_type,
            include=include,
            exclude=exclude,
            selector=selector,
            headers=headers,
            open_when_hidden=open_when_hidden,
            retry_interval_ms=retry_interval_ms,
            retry_scaler=retry_scaler,
            retry_max_wait_ms=retry_max_wait_ms,
            retry_max_count=retry_max_count,
            request_cancellation=request_cancellation,
        )

    def delete(
        self,
        uri: str,
        *,
        content_type: ContentType | str = "json",
        include: FilterValue | None = None,
        exclude: FilterValue | None = None,
        selector: str | None = None,
        headers: dict[str, str] | None = None,
        open_when_hidden: bool = False,
        retry_interval_ms: int = 1_000,
        retry_scaler: float = 2.0,
        retry_max_wait_ms: int = 30_000,
        retry_max_count: int = 10,
        request_cancellation: RequestCancellation | str = "auto",
    ) -> str:
        """
        Send a DELETE request to remove resources on the backend.

        Reference: https://data-star.dev/reference/actions#delete

        Args:
            uri: Endpoint URL or route name
            (See post() for full parameter documentation)

        Returns:
            Datastar @delete() action string

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.delete("/api/users/123")
            "@delete('/api/users/123')"
        """
        return self._http_action(
            "delete",
            uri,
            content_type=content_type,
            include=include,
            exclude=exclude,
            selector=selector,
            headers=headers,
            open_when_hidden=open_when_hidden,
            retry_interval_ms=retry_interval_ms,
            retry_scaler=retry_scaler,
            retry_max_wait_ms=retry_max_wait_ms,
            retry_max_count=retry_max_count,
            request_cancellation=request_cancellation,
        )

    # Pro Actions (require Datastar Pro)
    def clipboard(self, text: str, is_base64: bool = False) -> str:
        """
        Copy text to the system clipboard.

        Requires Datastar Pro. Copies text to the clipboard using the Clipboard API.

        Reference: https://data-star.dev/reference/actions#clipboard

        Args:
            text: Text to copy to clipboard
            is_base64: If True, text is base64 encoded

        Returns:
            Datastar @clipboard() action string

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.clipboard("Hello World")
            "@clipboard('Hello World')"
            >>> ds.clipboard("SGVsbG8=", is_base64=True)
            "@clipboard('SGVsbG8=', true)"
        """
        if is_base64:
            return f"@clipboard('{text}', true)"
        return f"@clipboard('{text}')"

    def fit(
        self,
        v: str,
        old_min: float,
        old_max: float,
        new_min: float,
        new_max: float,
        should_clamp: bool = False,
        should_round: bool = False,
    ) -> str:
        """
        Linearly interpolate a value from one range to another.

        Requires Datastar Pro. Maps a value from one numeric range to another,
        optionally clamping and rounding the result.

        Reference: https://data-star.dev/reference/actions#fit

        Args:
            v: Value expression to interpolate
            old_min: Minimum of original range
            old_max: Maximum of original range
            new_min: Minimum of target range
            new_max: Maximum of target range
            should_clamp: Clamp result to target range
            should_round: Round result to nearest integer

        Returns:
            Datastar @fit() action string

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.fit("$value", 0, 100, 0, 1)
            '@fit($value, 0, 100, 0, 1, false, false)'
            >>> ds.fit("$percent", 0, 100, 0, 255, should_clamp=True, should_round=True)
            '@fit($percent, 0, 100, 0, 255, true, true)'
        """
        return f"@fit({v}, {old_min}, {old_max}, {new_min}, {new_max}, {str(should_clamp).lower()}, {str(should_round).lower()})"

    @overload
    def fetch_on(
        self,
        event: Literal["intersect"],
        uri: str,
        *,
        # HTTP options
        method: Literal["get", "post", "put", "patch", "delete"] = "get",
        content_type: ContentType | str = "json",
        include: FilterValue | None = None,
        exclude: FilterValue | None = None,
        selector: str | None = None,
        headers: dict[str, str] | None = None,
        open_when_hidden: bool = False,
        retry_interval_ms: int = 1_000,
        retry_scaler: float = 2.0,
        retry_max_wait_ms: int = 30_000,
        retry_max_count: int = 10,
        request_cancellation: RequestCancellation | str = "auto",
        # intersect options
        once: bool = False,
        half: bool = False,
        full: bool = False,
        delay: TimeValue | None = None,
        debounce: Debounce | None = None,
        throttle: Throttle | None = None,
        viewtransition: bool = False,
    ) -> dict[str, str]:
        """
        Fetch a resource when the element intersects with the viewport.

        Convenience method that combines data-on-intersect with an HTTP action.
        Automatically triggers a fetch request when the element becomes visible.

        Reference: https://data-star.dev/reference/attributes#data-on-intersect
                   https://data-star.dev/reference/actions#backend-actions

        Args:
            event: Must be "intersect"
            uri: Endpoint URL or route name
            method: HTTP method to use
            content_type: Request content type
            include: Include only matching signals in request
            exclude: Exclude matching signals from request
            selector: CSS selector for elements to update
            headers: Additional HTTP headers
            open_when_hidden: Keep connection open when page is hidden
            retry_interval_ms: Initial retry delay
            retry_scaler: Exponential backoff multiplier
            retry_max_wait_ms: Maximum retry delay
            retry_max_count: Maximum retry attempts
            request_cancellation: Cancellation strategy
            once: Fire only once, then remove the listener
            half: Trigger when at least 50% is visible
            full: Trigger when 100% is visible
            delay: Delay before executing expression
            debounce: Debounce configuration
            throttle: Throttle configuration
            viewtransition: Use view transitions for updates

        Returns:
            Dictionary with data-on-intersect attribute

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.fetch_on("intersect", "/api/data")
            {'data-on-intersect': "@get('/api/data')"}
            >>> ds.fetch_on("intersect", "/api/users", method="post", once=True, half=True)
            {'data-on-intersect__once__half': "@post('/api/users')"}
        """

    @overload
    def fetch_on(
        self,
        event: Literal["interval"],
        uri: str,
        *,
        # HTTP options
        method: Literal["get", "post", "put", "patch", "delete"] = "get",
        content_type: ContentType | str = "json",
        include: FilterValue | None = None,
        exclude: FilterValue | None = None,
        selector: str | None = None,
        headers: dict[str, str] | None = None,
        open_when_hidden: bool = False,
        retry_interval_ms: int = 1_000,
        retry_scaler: float = 2.0,
        retry_max_wait_ms: int = 30_000,
        retry_max_count: int = 10,
        request_cancellation: RequestCancellation | str = "auto",
        # interval options
        duration: TimeValue | tuple[TimeValue, Literal["leading"]] = "1s",
        viewtransition: bool = False,
    ) -> dict[str, str]:
        """
        Fetch a resource at regular intervals.

        Convenience method that combines data-on-interval with an HTTP action.
        Automatically triggers fetch requests at the specified interval.

        Reference: https://data-star.dev/reference/attributes#data-on-interval
                   https://data-star.dev/reference/actions#backend-actions

        Args:
            event: Must be "interval"
            uri: Endpoint URL or route name
            method: HTTP method to use
            content_type: Request content type
            include: Include only matching signals in request
            exclude: Exclude matching signals from request
            selector: CSS selector for elements to update
            headers: Additional HTTP headers
            open_when_hidden: Keep connection open when page is hidden
            retry_interval_ms: Initial retry delay
            retry_scaler: Exponential backoff multiplier
            retry_max_wait_ms: Maximum retry delay
            retry_max_count: Maximum retry attempts
            request_cancellation: Cancellation strategy
            duration: Interval duration or tuple with duration and "leading" modifier
            viewtransition: Use view transitions for updates

        Returns:
            Dictionary with data-on-interval attribute

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.fetch_on("interval", "/api/data")
            {'data-on-interval': "@get('/api/data')"}
            >>> ds.fetch_on("interval", "/api/poll", duration=5)
            {'data-on-interval__duration.5s': "@get('/api/poll')"}
            >>> ds.fetch_on("interval", "/api/check", duration=(2, "leading"))
            {'data-on-interval__duration.2s.leading': "@get('/api/check')"}
        """

    @overload
    def fetch_on(
        self,
        event: Literal["load"],
        uri: str,
        *,
        # HTTP options
        method: Literal["get", "post", "put", "patch", "delete"] = "get",
        content_type: ContentType | str = "json",
        include: FilterValue | None = None,
        exclude: FilterValue | None = None,
        selector: str | None = None,
        headers: dict[str, str] | None = None,
        open_when_hidden: bool = False,
        retry_interval_ms: int = 1_000,
        retry_scaler: float = 2.0,
        retry_max_wait_ms: int = 30_000,
        retry_max_count: int = 10,
        request_cancellation: RequestCancellation | str = "auto",
        # load options
        delay: TimeValue | None = None,
        viewtransition: bool = False,
    ) -> dict[str, str]:
        """
        Fetch a resource when the element loads.

        Convenience method that combines data-on-load with an HTTP action.
        Automatically triggers a fetch request when the element appears.

        Reference: https://data-star.dev/reference/attributes#data-on-load
                   https://data-star.dev/reference/actions#backend-actions

        Args:
            event: Must be "load"
            uri: Endpoint URL or route name
            method: HTTP method to use
            content_type: Request content type
            include: Include only matching signals in request
            exclude: Exclude matching signals from request
            selector: CSS selector for elements to update
            headers: Additional HTTP headers
            open_when_hidden: Keep connection open when page is hidden
            retry_interval_ms: Initial retry delay
            retry_scaler: Exponential backoff multiplier
            retry_max_wait_ms: Maximum retry delay
            retry_max_count: Maximum retry attempts
            request_cancellation: Cancellation strategy
            delay: Delay before fetching
            viewtransition: Use view transitions for updates

        Returns:
            Dictionary with data-on-load attribute

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.fetch_on("load", "/api/data")
            {'data-on-load': "@get('/api/data')"}
            >>> ds.fetch_on("load", "/api/users", method="post", delay=1)
            {'data-on-load__delay.1s': "@post('/api/users')"}
        """

    @overload
    def fetch_on(
        self,
        event: Literal["signal-patch"],
        uri: str,
        *,
        # HTTP options
        method: Literal["get", "post", "put", "patch", "delete"] = "get",
        content_type: ContentType | str = "json",
        include: FilterValue | None = None,
        exclude: FilterValue | None = None,
        selector: str | None = None,
        headers: dict[str, str] | None = None,
        open_when_hidden: bool = False,
        retry_interval_ms: int = 1_000,
        retry_scaler: float = 2.0,
        retry_max_wait_ms: int = 30_000,
        retry_max_count: int = 10,
        request_cancellation: RequestCancellation | str = "auto",
        # signal-patch options
        delay: TimeValue | None = None,
        debounce: Debounce | None = None,
        throttle: Throttle | None = None,
        # Filter
        signal_include: FilterValue | None = None,
        signal_exclude: FilterValue | None = None,
    ) -> dict[str, str]:
        """
        Fetch a resource when signals are patched from the backend.

        Convenience method that combines data-on-signal-patch with an HTTP action.
        Automatically triggers fetch requests when server-sent events update signals.

        Reference: https://data-star.dev/reference/attributes#data-on-signal-patch
                   https://data-star.dev/reference/actions#backend-actions

        Args:
            event: Must be "signal-patch"
            uri: Endpoint URL or route name
            method: HTTP method to use
            content_type: Request content type
            include: Include only matching signals in request
            exclude: Exclude matching signals from request
            selector: CSS selector for elements to update
            headers: Additional HTTP headers
            open_when_hidden: Keep connection open when page is hidden
            retry_interval_ms: Initial retry delay
            retry_scaler: Exponential backoff multiplier
            retry_max_wait_ms: Maximum retry delay
            retry_max_count: Maximum retry attempts
            request_cancellation: Cancellation strategy
            delay: Delay before executing expression
            debounce: Debounce configuration
            throttle: Throttle configuration
            signal_include: Only trigger for signals matching this pattern
            signal_exclude: Don't trigger for signals matching this pattern

        Returns:
            Dictionary with data-on-signal-patch attribute(s)

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.fetch_on("signal-patch", "/api/data")
            {'data-on-signal-patch': "@get('/api/data')"}
            >>> ds.fetch_on("signal-patch", "/api/log", signal_include="user.*", debounce=0.5)
            {'data-on-signal-patch__debounce.500ms': "@get('/api/log')", 'data-on-signal-patch-filter': '{"include":"user.*"}'}
        """

    @overload
    def fetch_on(
        self,
        event: JSEvent | str,
        uri: str,
        *,
        # HTTP options
        method: Literal["get", "post", "put", "patch", "delete"] = "get",
        content_type: ContentType | str = "json",
        include: FilterValue | None = None,
        exclude: FilterValue | None = None,
        selector: str | None = None,
        headers: dict[str, str] | None = None,
        open_when_hidden: bool = False,
        retry_interval_ms: int = 1_000,
        retry_scaler: float = 2.0,
        retry_max_wait_ms: int = 30_000,
        retry_max_count: int = 10,
        request_cancellation: RequestCancellation | str = "auto",
        # event options
        once: bool = False,
        passive: bool = False,
        capture: bool = False,
        delay: TimeValue | None = None,
        debounce: Debounce | None = None,
        throttle: Throttle | None = None,
        viewtransition: bool = False,
        window: bool = False,
        outside: bool = False,
        prevent: bool = False,
        stop: bool = False,
    ) -> dict[str, str]:
        """
        Fetch a resource in response to a DOM event.

        Convenience method that combines data-on-{event} with an HTTP action.
        Automatically triggers fetch requests when the specified DOM event fires.

        Reference: https://data-star.dev/reference/attributes#data-on
                   https://data-star.dev/reference/actions#backend-actions

        Args:
            event: Event name (click, input, submit, etc.) or custom event name
            uri: Endpoint URL or route name
            method: HTTP method to use
            content_type: Request content type
            include: Include only matching signals in request
            exclude: Exclude matching signals from request
            selector: CSS selector for elements to update
            headers: Additional HTTP headers
            open_when_hidden: Keep connection open when page is hidden
            retry_interval_ms: Initial retry delay
            retry_scaler: Exponential backoff multiplier
            retry_max_wait_ms: Maximum retry delay
            retry_max_count: Maximum retry attempts
            request_cancellation: Cancellation strategy
            once: Fire only once, then remove the listener
            passive: Mark listener as passive (improves scroll performance)
            capture: Use capture phase instead of bubble phase
            delay: Delay before executing expression
            debounce: Debounce configuration
            throttle: Throttle configuration
            viewtransition: Use view transitions for updates
            window: Attach listener to window instead of element
            outside: Trigger when event occurs outside the element
            prevent: Call preventDefault() on the event
            stop: Call stopPropagation() on the event

        Returns:
            Dictionary with data-on-{event} attribute

        Examples:
            >>> ds = DatastarAttributes(Stario())
            >>> ds.fetch_on("click", "/api/data")
            {'data-on-click': "@get('/api/data')"}
            >>> ds.fetch_on("submit", "/api/save", method="post", prevent=True)
            {'data-on-submit__prevent': "@post('/api/save')"}
            >>> ds.fetch_on("input", "/api/search", debounce=0.3)
            {'data-on-input__debounce.300ms': "@get('/api/search')"}
            >>> ds.fetch_on("click", "/api/close", outside=True)
            {'data-on-click__outside': "@get('/api/close')"}
            >>> ds.fetch_on("scroll", "/api/track", window=True, throttle=0.1)
            {'data-on-scroll__window__throttle.100ms': "@get('/api/track')"}
        """

    def fetch_on(
        self,
        event: str,
        uri: str,
        # HTTP options
        method: Literal["get", "post", "put", "patch", "delete"] = "get",
        content_type: ContentType | str = "json",
        include: FilterValue | None = None,
        exclude: FilterValue | None = None,
        selector: str | None = None,
        headers: dict[str, str] | None = None,
        open_when_hidden: bool = False,
        retry_interval_ms: int = 1_000,
        retry_scaler: float = 2.0,
        retry_max_wait_ms: int = 30_000,
        retry_max_count: int = 10,
        request_cancellation: RequestCancellation | str = "auto",
        **kwargs: Any,
    ) -> dict[str, str]:
        """
        Fetch a resource in response to an event.

        Convenience method that combines data-on-{event} with an HTTP action.
        Automatically triggers fetch requests when the specified event fires.

        This method works with all event types:
        - DOM events: "click", "input", "submit", etc.
        - Special events: "intersect", "interval", "load", "signal-patch"

        Reference: https://data-star.dev/reference/attributes#data-on
                   https://data-star.dev/reference/actions#backend-actions
        """
        # Generate HTTP action expression
        expression = self._http_action(
            method,
            uri,
            content_type=content_type,
            include=include,
            exclude=exclude,
            selector=selector,
            headers=headers,
            open_when_hidden=open_when_hidden,
            retry_interval_ms=retry_interval_ms,
            retry_scaler=retry_scaler,
            retry_max_wait_ms=retry_max_wait_ms,
            retry_max_count=retry_max_count,
            request_cancellation=request_cancellation,
        )

        # Route to appropriate handler based on event type
        if event == "intersect":
            return self._on_intersect(expression, **kwargs)
        if event == "interval":
            return self._on_interval(expression, **kwargs)
        if event == "load":
            return self._on_load(expression, **kwargs)
        if event == "signal-patch":
            return self._on_signal_patch(expression, **kwargs)
        return self._on_event(event, expression, **kwargs)


type Datastar = Annotated[DatastarAttributes, DatastarAttributes, "singleton"]
"""
Type alias for the DatastarAttributes singleton dependency.

Use this type for dependency injection or type hints when you want to access
the DatastarAttributes generator, which provides methods to build and manage
data-* attributes and Datastar actions for your HTML elements.

Example (building an HTML element with Datastar attributes and actions):

    from stario.html import button
    def my_button(ds: Datastar):
        return button(
            ds.bind("buttonLabel"),
            ds.text("$buttonLabel"),
            ds.on("click", f'ds.get("/api/do-something") && $buttonLabel = $buttonLabel + " clicked"'),
            "Click me",
        )

This type alias ensures consistent usage and discoverability throughout your
application and extensions.
"""



from .loggers.route import Logger as Logger
from .loggers.route import RouteLogger as RouteLogger
from .queue import LogQueue as LogQueue
from .sinks.json import JSONSink as JSONSink
from .sinks.rich import RichConsoleSink as RichConsoleSink
from .types import LogLevel as LogLevel



from dataclasses import dataclass
from enum import IntEnum
from typing import Any, Protocol


class LogLevel(IntEnum):
    """Log levels with numeric values for comparison."""

    DEBUG = 10
    INFO = 20
    WARNING = 30
    ERROR = 40
    CRITICAL = 50


@dataclass(slots=True)
class BaseLogRecord:
    """Base log entry."""

    timestamp: float  # UTC timestamp
    request_id: str  # UUIDv4 correlation ID


@dataclass(slots=True)
class RouteLogRecord(BaseLogRecord):
    """Application/debug log entry."""

    level: LogLevel
    message: str
    context: dict[str, Any]
    buffered: bool = False  # Sent only if log level is above threshold
    background: bool = False  # Sent from background tasks (after response has started)
    exc_info: tuple[type, BaseException, Any] | None = None


@dataclass(slots=True)
class RequestLogRecord(BaseLogRecord):
    """HTTP request log entry."""

    method: str
    path: str
    client: tuple[str, int] | None  # (host, port)


@dataclass(slots=True)
class ResponseLogRecord(BaseLogRecord):
    """HTTP response log entry."""

    status_code: int
    exc_info: tuple[type, BaseException, Any] | None


# Union type for all log records
Record = RequestLogRecord | ResponseLogRecord | RouteLogRecord


class Sink(Protocol):
    """Protocol for log output destinations."""

    def open(self) -> None:
        """Open the sink."""
        ...

    def close(self) -> None:
        """Close the sink."""
        ...

    async def write(self, records: list[Record]) -> None:
        """Write one or more log records to the sink."""
        ...


class LogQueue(Protocol):
    """Protocol for a log queue."""

    def enqueue(self, record: Record) -> None: ...



import sys
import uuid

from starlette._utils import is_async_callable
from starlette.concurrency import run_in_threadpool
from starlette.requests import Request
from starlette.responses import PlainTextResponse
from starlette.types import ASGIApp, ExceptionHandler, Message, Receive, Scope, Send

from stario.logging.loggers.access import AccessLogger
from stario.logging.queue import LogQueue


class GuardianMiddleware:
    """
    Handles returning 500 responses when a server error occurs.

    If 'debug' is set, then traceback responses will be returned,
    otherwise the designated 'handler' will be called.

    This middleware class should generally be used to wrap *everything*
    else up, so that unhandled exceptions anywhere in the stack
    always result in an appropriate 500 response.

    Based on starlette.middleware.exceptions.ExceptionMiddleware
    """

    def __init__(
        self,
        app: ASGIApp,
        log_queue: LogQueue,
        handler: ExceptionHandler | None = None,
    ) -> None:
        self.app = app
        self.handler = handler
        self.log_queue = log_queue
        self.access_logger = AccessLogger(log_queue)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        # We need uid and duration
        request_id = str(uuid.uuid4())
        scope["stario.request_id"] = request_id
        exception_info = None

        # This is to track if the response has started
        #  borrowed from starlette
        response_started = False
        status_code = None

        async def _send(message: Message) -> None:
            nonlocal response_started, status_code, send

            if message["type"] == "http.response.start":
                response_started = True
                status_code = message["status"]

                # Update scope with useful information
                scope["stario.response_started"] = True
                message["headers"].append(
                    (b"x-request-id", request_id.encode("latin-1"))
                )

            # Continue sending the message
            await send(message)

        self.access_logger.request(
            request_id=request_id,
            method=scope["method"],
            path=scope["path"],
            client=scope.get("client"),
        )

        try:
            await self.app(scope, receive, _send)
        except Exception as exc:
            request = Request(scope)

            if self.handler is None:
                # Use our default 500 error handler.
                response = PlainTextResponse("Internal Server Error", status_code=500)

                exception_info = sys.exc_info()
                if exception_info == (None, None, None):
                    exception_info = None
                else:
                    # Ensure exc_info is properly typed
                    exception_info = (
                        exception_info if exception_info[0] is not None else None
                    )
            else:
                # Use an installed 500 error handler.
                if is_async_callable(self.handler):
                    # TODO: this needs a bit of typing magic fixes:
                    response = await self.handler(request, exc)  # type: ignore[assignment]
                else:
                    response = await run_in_threadpool(self.handler, request, exc)

            if not response_started and response is not None:
                # Response objects are callable ASGI applications
                await response(scope, receive, _send)

            # We always continue to raise the exception.
            # This allows servers to log the error, or allows test clients
            # to optionally raise the error within the test case.
            # raise exc

        # Log the response info
        self.access_logger.response(
            request_id=request_id,
            status_code=status_code or 500,
            exc_info=exception_info,
        )



import asyncio
import sys
from typing import Iterable

from stario.logging.sinks.json import JSONSink
from stario.logging.sinks.rich import RichConsoleSink
from stario.logging.types import (
    Record,
    Sink,
)


class LogQueue:
    """
    Log queue that handles buffering and flushing to sinks.

    Uses a listener pattern where loggers call enqueue() which handles
    buffering and delayed flushing via asyncio tasks.
    """

    def __init__(
        self,
        sinks: Iterable[Sink] | None = None,
        buffer_size: int = 100,
        flush_interval: float = 0.05,
    ):

        if sinks is None:
            if sys.stdout.isatty():
                self.sinks = [RichConsoleSink()]
            else:
                self.sinks = [JSONSink()]
        else:
            self.sinks = list(sinks)

        self.buffer_size = buffer_size
        self.flush_interval = flush_interval

        self._buffer: list[Record] = []
        self._flush_task: asyncio.Task | None = None
        self._running = False

    def enqueue(self, record: Record) -> None:
        """
        Enqueue a log record for writing.

        Implements throttled buffering:
        - If buffer full: flush immediately
        - Otherwise: buffer and ensure flush task is running
        - Flush task runs at regular intervals (throttling, not debouncing)
        """
        self._buffer.append(record)

        # If buffer is full, schedule immediate flush
        if len(self._buffer) >= self.buffer_size:
            if self._flush_task and not self._flush_task.done():
                # Cancel current flush task and create new one
                self._flush_task.cancel()
            self._flush_task = asyncio.create_task(self._flush_buffer())

    async def _buffer_flusher(self) -> None:
        """
        Flush the buffer to all sinks at regular intervals.
        """
        try:
            while self._running:
                await self._flush_buffer()

                await asyncio.sleep(self.flush_interval)

            # Flush the buffer one last time before stopping
            await self._flush_buffer()

            for sink in self.sinks:
                sink.close()

        except Exception as e:
            print(f"Error in buffer flusher: {type(e).__name__}: {str(e)}")
            raise

    async def _flush_buffer(self) -> None:
        """Flush the current buffer to all sinks."""
        if not self._buffer:
            return

        # Swap buffers
        records = self._buffer
        self._buffer = []

        # Write to all sinks; if only one, do it directly for efficiency
        if len(self.sinks) == 1:
            await self.sinks[0].write(records)

        else:
            await asyncio.gather(
                *[sink.write(records) for sink in self.sinks],
                return_exceptions=True,
            )

    def start(self) -> None:
        """Start the log queue."""
        self._running = True
        self._flush_task = asyncio.create_task(self._buffer_flusher())

        for sink in self.sinks:
            sink.open()

    def stop(self) -> None:
        """Stop and flush remaining logs."""
        self._running = False



from typing import Any, Callable, Sequence, TypeVar

from starlette.datastructures import URLPath
from starlette.middleware import Middleware
from starlette.routing import BaseRoute, Router
from starlette.types import ASGIApp, Lifespan, Receive, Scope, Send

_T = TypeVar("_T")


class _DefaultLifespan:
    def __init__(self, router: "StarRouter"):
        self._router = router

    async def __aenter__(self) -> None:
        await self._router.startup()

    async def __aexit__(self, *exc_info: object) -> None:
        await self._router.shutdown()

    def __call__(self: _T, app: object) -> _T:
        return self


class StarRouter:
    def __init__(
        self,
        *routes: BaseRoute,
        redirect_slashes: bool = True,
        default: ASGIApp | None = None,
        lifespan: Lifespan[Any] | None = None,
        middleware: Sequence[Middleware] | None = None,
    ) -> None:

        # This is basically borrowed from Starlette's Router
        self.routes = list(routes)
        self.redirect_slashes = redirect_slashes
        self.default = self.not_found if default is None else default
        self.on_startup: list[Callable[[], Any]] = []
        self.on_shutdown: list[Callable[[], Any]] = []

        self.lifespan_context = lifespan or _DefaultLifespan(self)

        self.middleware_stack = self.app
        if middleware:
            for cls, args, kwargs in reversed(middleware):
                self.middleware_stack = cls(self.middleware_stack, *args, **kwargs)

    async def not_found(self, scope: Scope, receive: Receive, send: Send) -> None:
        await Router.not_found(self, scope, receive, send)  # type: ignore[arg-type]

    def url_path_for(self, name: str, /, **path_params: Any) -> URLPath:
        # Starlette Router depends here only on self.routes
        return Router.url_path_for(self, name, **path_params)  # type: ignore[arg-type]

    async def startup(self) -> None:
        # Starlette Router depends here only on self.on_startup
        await Router.startup(self)  # type: ignore[arg-type]

    async def shutdown(self) -> None:
        # Starlette Router depends here only on self.on_shutdown
        await Router.shutdown(self)  # type: ignore[arg-type]

    async def lifespan(self, scope: Scope, receive: Receive, send: Send) -> None:
        # Starlette Router depends here only on self.lifespan_context
        await Router.lifespan(self, scope, receive, send)  # type: ignore[arg-type]

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        await self.middleware_stack(scope, receive, send)

    async def app(self, scope: Scope, receive: Receive, send: Send) -> None:
        # Starlette Router depends here only on self.lifespan(), self.routes and self.default
        await Router.app(self, scope, receive, send)  # type: ignore[arg-type]

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, StarRouter) and self.routes == other.routes

    def mount(self, path: str, app: ASGIApp, name: str | None = None) -> None:
        Router.mount(self, path, app=app, name=name)  # type: ignore[arg-type]

    def host(self, host: str, app: ASGIApp, name: str | None = None) -> None:
        Router.host(self, host, app=app, name=name)  # type: ignore[arg-type]

    def add(self, route: BaseRoute) -> None:
        self.routes.append(route)

    def add_event_handler(self, event_type: str, func: Callable[[], Any]) -> None:
        Router.add_event_handler(self, event_type, func)  # type: ignore[arg-type]



from starlette.responses import FileResponse as FileResponse
from starlette.responses import HTMLResponse as HTMLResponse
from starlette.responses import JSONResponse as JSONResponse
from starlette.responses import PlainTextResponse as PlainTextResponse
from starlette.responses import RedirectResponse as RedirectResponse
from starlette.responses import Response as Response
from starlette.responses import StreamingResponse as StreamingResponse



import json
from dataclasses import dataclass
from typing import (
    AsyncGenerator,
    Callable,
    Final,
    Literal,
    cast,
    get_args,
)

from stario.html import Attributes, render, script

type PatchMode = Literal[
    "outer",
    "inner",
    "replace",
    "prepend",
    "append",
    "before",
    "after",
    "remove",
]
PatchModes: Final[frozenset[PatchMode]] = frozenset(get_args(PatchMode))
type PatchSelector = str
type SignalValue = str | int | float | bool | list[SignalValue] | dict[str, SignalValue]
type SignalsDict = dict[str, SignalValue]

# Server-Sent Events (SSE) related
type PatchSignalsEvent = SignalsPatch | SignalsDict
type PatchElementsEvent[ElementT] = (
    ElementsPatch[ElementT]
    | ElementT
    | tuple[PatchMode, PatchSelector, ElementT]
    | tuple[PatchMode, ElementT]
    | tuple[Literal["remove"], PatchSelector]
    | tuple[Literal["script"], str]
    | tuple[Literal["redirect"], str]
)
type Event[ElementT] = (
    PatchElementsEvent[ElementT] | PatchSignalsEvent | ScriptExecution | Redirection
)
type EventStream[ElementT] = AsyncGenerator[Event[ElementT], None]


@dataclass(slots=True)
class ElementsPatch[ElementT]:
    """
    https://data-star.dev/reference/sse_events#datastar-patch-elements
    """

    mode: PatchMode = "outer"
    selector: PatchSelector | None = None
    elements: ElementT | None = None
    use_view_transition: bool = False

    # SSE specific
    event_id: str | None = None
    retry_duration: int | None = None

    def to_sse(self, renderer: Callable[[ElementT], str]) -> str:
        # Standard SSE headers
        lines = ["event: datastar-patch-elements"]
        append = lines.append

        if self.event_id:
            append(f"id: {self.event_id}")

        if self.retry_duration and self.retry_duration != 1000:
            append(f"retry: {self.retry_duration}")

        # Datastar specific:
        if self.mode and self.mode != "outer":
            append(f"data: mode {self.mode}")

        if self.selector:
            append(f"data: selector {self.selector}")

        if self.use_view_transition:
            append(f"data: useViewTransition {self.use_view_transition}")

        if self.elements:

            # Render the element using the renderer
            element = renderer(self.elements)

            # Split elements into lines - this should be faster than splitlines()
            start = 0
            while True:
                end = element.find("\n", start)
                if end == -1:
                    append(f"data: elements {element[start:]}")
                    break
                append(f"data: elements {element[start:end]}")
                start = end + 1

        return "\n".join(lines) + "\n\n"


@dataclass(slots=True)
class SignalsPatch:
    """
    https://data-star.dev/reference/sse_events#datastar-patch-signals
    """

    signals: SignalsDict
    only_if_missing: bool = False

    # SSE specific
    event_id: str | None = None
    retry_duration: int | None = None

    def to_sse(self) -> str:

        # Standard SSE headers
        lines = ["event: datastar-patch-signals"]
        append = lines.append

        if self.event_id:
            append(f"id: {self.event_id}")

        if self.retry_duration and self.retry_duration != 1000:
            append(f"retry: {self.retry_duration}")

        if self.only_if_missing:
            js_bool = "true" if self.only_if_missing else "false"
            append(f"data: onlyIfMissing {js_bool}")

        # Datastar specific:
        # Split json into lines - this should be faster than splitlines()
        json_str = json.dumps(self.signals)
        start = 0
        while True:
            end = json_str.find("\n", start)
            if end == -1:
                append(f"data: signals {json_str[start:]}")
                break
            append(f"data: signals {json_str[start:end]}")
            start = end + 1

        return "\n".join(lines) + "\n\n"


@dataclass(slots=True)
class ElementsRemoval[ElementT]:

    selector: PatchSelector | None = None
    elements: ElementT | None = None

    # SSE specific
    event_id: str | None = None
    retry_duration: int | None = None

    def to_sse(self, renderer: Callable[[ElementT], str]) -> str:
        return ElementsPatch(
            mode="remove",
            selector=self.selector,
            elements=self.elements,
            event_id=self.event_id,
            retry_duration=self.retry_duration,
        ).to_sse(renderer)


@dataclass(slots=True)
class ScriptExecution:

    script: str
    auto_remove: bool = True
    attributes: Attributes | None = None

    # SSE specific
    event_id: str | None = None
    retry_duration: int | None = None

    def to_sse(self) -> str:

        return ElementsPatch(
            mode="append",
            selector="body",
            elements=script(
                self.attributes or {},
                {"data-effect": "el.remove();"} if self.auto_remove else {},
                self.script,
            ),
            event_id=self.event_id,
            retry_duration=self.retry_duration,
        ).to_sse(render)


@dataclass(slots=True)
class Redirection:

    location: str

    # SSE specific
    event_id: str | None = None
    retry_duration: int | None = None

    def to_sse(self) -> str:
        script = f"setTimeout(() => window.location = '{self.location}');"
        return ScriptExecution(
            script=script,
            event_id=self.event_id,
            retry_duration=self.retry_duration,
        ).to_sse()


def patch_to_sse[ElementT](
    patch: Event[ElementT],
    renderer: Callable[[ElementT], str],
) -> str:

    if isinstance(patch, (SignalsPatch, ScriptExecution, Redirection)):
        # yield SignalsPatch or ScriptExecution or Redirection directly
        return patch.to_sse()

    if isinstance(patch, (ElementsPatch, ElementsRemoval)):
        # yield ElementsPatch or ElementsRemoval directly
        return patch.to_sse(renderer)

    if isinstance(patch, dict):
        # yield {"signal1": "value1", "signal2": "value2"}
        return SignalsPatch(signals=patch).to_sse()

    if isinstance(patch, tuple):

        if len(patch) == 3 and patch[0] in PatchModes:
            # yield (PatchMode, PatchSelector, ElementT)
            return ElementsPatch(
                mode=patch[0], selector=patch[1], elements=patch[2]
            ).to_sse(renderer)

        if len(patch) == 2:

            first, second = patch

            if first == "script" and isinstance(second, str):
                # yield ("script", str)
                return ScriptExecution(script=second).to_sse()

            if first == "redirect" and isinstance(second, str):
                # yield ("redirect", str)
                return Redirection(location=second).to_sse()

            if first in PatchModes:
                # yield ("remove", PatchSelector)
                # or yield (PatchMode, ElementT)

                if (
                    first == "remove"
                    and isinstance(second, str)
                    and not (second.startswith("<") and second.endswith(">"))
                ):
                    # Special case when removing by providing a selector string (not HTML elements)
                    return ElementsRemoval(selector=second).to_sse(renderer)

                # yield (PatchMode, ElementT)
                second = cast(ElementT, second)
                return ElementsPatch(mode=first, elements=second).to_sse(renderer)

    # yield (elements) # HTML elements
    # We assume anything else is a HTML element that will be rendered by provided renderer
    return ElementsPatch[ElementT](elements=cast(ElementT, patch)).to_sse(renderer)



from brotli_asgi import BrotliMiddleware as AsgiBrotliMiddleware
from starlette.middleware import Middleware
from starlette.types import ASGIApp


class BrotliMiddleware(AsgiBrotliMiddleware):

    def __init__(
        self,
        app: ASGIApp,
        quality: int = 4,
        mode: str = "text",
        lgwin: int = 22,
        lgblock: int = 0,
        minimum_size: int = 400,
        gzip_fallback: bool = True,
        excluded_handlers: list | None = None,
    ) -> None:
        """
        Args:
            mode: The compression mode can be:
                generic, text (*default*. Used for UTF-8 format text input)
                or font (for WOFF 2.0).
            quality: Controls the compression-speed vs compression-
                density tradeoff. The higher the quality, the slower the compression.
                Range is 0 to 11.
            lgwin: Base 2 logarithm of the sliding window size. Range
                is 10 to 24.
            lgblock: Base 2 logarithm of the maximum input block size.
                Range is 16 to 24. If set to 0, the value will be set based on the
                quality.
            minimum_size: Only compress responses that are bigger than this value in bytes.
            gzip_fallback: If True, uses gzip encoding if br is not in the Accept-Encoding header.
            excluded_handlers: List of handlers to be excluded from being compressed.
        """
        # fmt: off
        super().__init__(
            app               = app,
            quality           = quality,
            mode              = mode,
            lgwin             = lgwin,
            lgblock           = lgblock,
            minimum_size      = minimum_size,
            gzip_fallback     = gzip_fallback,
            excluded_handlers = excluded_handlers,
        )
        # fmt: on

    @classmethod
    def as_middleware(
        cls,
        quality: int = 4,
        mode: str = "text",
        lgwin: int = 22,
        lgblock: int = 0,
        minimum_size: int = 400,
        gzip_fallback: bool = True,
        excluded_handlers: list | None = None,
    ) -> Middleware:

        # fmt: off
        return Middleware(
            cls,
            quality           = quality,
            mode              = mode,
            lgwin             = lgwin,
            lgblock           = lgblock,
            minimum_size      = minimum_size,
            gzip_fallback     = gzip_fallback,
            excluded_handlers = excluded_handlers,
        )
        # fmt: on



"""
Type definitions and aliases for Stario.

This module contains shared type definitions that are used across multiple modules.
By centralizing these types here, we avoid circular import issues.
"""

from typing import Awaitable, Callable, Mapping, Sequence

from starlette.types import Receive, Scope, Send

# Headers constraint type - used for route header matching
type HeadersConstraint = Mapping[str, str | None] | Sequence[
    str | tuple[str, str | None]
]

# Endpoint function type
type EndpointFunction[T] = Callable[..., T]

# Request handler type
type RequestHandler = Callable[[Scope, Receive, Send], Awaitable[None]]

# Adapter function type
type AdapterFunction[T] = Callable[[EndpointFunction[T]], RequestHandler]



from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import Dict, List, Optional

from rich.console import Group, RenderableType
from rich.live import Live
from rich.panel import Panel
from rich.rule import Rule
from rich.text import Text
from rich.traceback import Traceback

from stario.logging.types import (
    LogLevel,
    Record,
    RequestLogRecord,
    ResponseLogRecord,
    RouteLogRecord,
)


@dataclass
class RequestGroup:
    """Tracks all logs for a single request by request_id."""

    request_id: str
    start_timestamp: float
    request_log: Optional[RequestLogRecord] = None
    route_logs: List[RouteLogRecord] = field(default_factory=list)
    response_log: Optional[ResponseLogRecord] = None

    @property
    def duration_ms(self) -> float:
        """Calculate total request duration in milliseconds."""
        if not self.response_log:
            return 0.0
        return (self.response_log.timestamp - self.start_timestamp) * 1000


class RichConsoleSink:
    """Rich console output with request grouping and live updates."""

    # Color schemes
    LOG_LEVEL_COLORS = {
        LogLevel.DEBUG: "cyan",
        LogLevel.INFO: "green",
        LogLevel.WARNING: "yellow",
        LogLevel.ERROR: "red",
        LogLevel.CRITICAL: "magenta",
    }

    HTTP_METHOD_COLORS = {
        "GET": "blue",
        "POST": "green",
        "PUT": "yellow",
        "DELETE": "red",
        "PATCH": "magenta",
        "HEAD": "cyan",
        "OPTIONS": "white",
    }

    STATUS_CODE_COLORS = {
        "2xx": "green",
        "3xx": "yellow",
        "4xx": "red",
        "5xx": "bright_red",
    }

    def __init__(self, max_live_requests: int = 100):
        # self.console = Console()
        self.max_live_requests = max_live_requests
        self.request_groups: Dict[str, RequestGroup] = {}  # In-progress requests only
        self.live = Live(auto_refresh=False)
        self._started = False

    def open(self) -> None:
        """Open the sink."""
        self.live.start()
        self._started = True

    def close(self) -> None:
        """Close the sink."""
        if self._started:
            if self.live:
                self.live.stop()
            self._started = False

    async def write(self, records: list[Record]) -> None:
        """Write formatted log records with request grouping."""
        # Process all records
        for record in records:
            self._process_record(record)

        # Update live display only if there are in-progress requests
        # if self.request_groups:
        content = self._create_live_content()
        self.live.update(content, refresh=True)

    def _process_record(self, record: Record) -> None:
        """Process a single record and update request groups."""
        request_id = record.request_id

        # Get or create request group
        if request_id not in self.request_groups:
            self.request_groups[request_id] = RequestGroup(
                request_id=request_id, start_timestamp=record.timestamp
            )

        group = self.request_groups[request_id]
        if isinstance(record, RequestLogRecord):
            group.request_log = record
        elif isinstance(record, RouteLogRecord):
            group.route_logs.append(record)
        elif isinstance(record, ResponseLogRecord):
            group.response_log = record
            # Render completed request statically
            self._render_completed_request(group)
            # Remove from in-progress tracking
            self.request_groups.pop(request_id)

    def _render_completed_request(self, group: RequestGroup) -> None:
        """Render a completed request statically to console."""
        if group.route_logs or (group.response_log and group.response_log.exc_info):
            panel = self._create_request_panel(group)
            self.live.console.print(panel)
        else:
            line = self._create_single_line(group)
            self.live.console.print(line)

    def _create_live_content(self) -> Group:
        """Create content for live display - sorted in-progress requests."""

        if not self.request_groups:
            return Group()

        content_parts = []

        # Sort in-progress by start timestamp and limit to max_live_requests
        in_progress_groups = sorted(
            self.request_groups.values(), key=lambda g: g.start_timestamp
        )

        # Limit to max_live_requests (show oldest ones)
        if len(in_progress_groups) > self.max_live_requests:
            in_progress_groups = in_progress_groups[: self.max_live_requests]

        for group in in_progress_groups:
            if group.route_logs or (group.response_log and group.response_log.exc_info):
                panel = self._create_request_panel(group)
                content_parts.append(panel)
            else:
                line = self._create_single_line(group)
                content_parts.append(line)

        # Return empty Group if no content to avoid rendering empty Text
        return Group(*content_parts)

    def _create_request_panel(self, group: RequestGroup) -> Panel:
        """Create a Rich panel for a request group."""
        # Create panel title
        title = self._create_panel_title(group)

        # Create panel content (route logs only)
        content = self._create_panel_content(group)

        # Determine panel style based on completion status
        if group.response_log is not None:
            border_style = self._get_status_color(group.response_log.status_code)
        else:
            border_style = "dim white"

        return Panel(
            content,
            title=title,
            title_align="left",
            border_style=border_style,
            padding=(0, 1),
        )

    def _create_panel_title(self, group: RequestGroup) -> Text:
        """Create the panel title matching single-line format."""
        title = Text()

        # Timestamp
        dt = datetime.fromtimestamp(group.start_timestamp, tz=timezone.utc)
        time_str = dt.strftime("%H:%M:%S.%f")[:-3]
        title.append(f"{time_str} | ", style="dim")

        # Request ID (first 8 chars)
        title.append(f"{group.request_id[:8]} | ", style="dim")

        if group.request_log:
            # HTTP method (padded to 7 chars)
            method_color = self.HTTP_METHOD_COLORS.get(
                group.request_log.method, "white"
            )
            title.append(
                f"{group.request_log.method:<7} ", style=f"bold {method_color}"
            )

            # Path (padded to 28 characters for alignment)
            title.append(f"{group.request_log.path:<50} ", style="white")

        if group.response_log is not None:
            # Status code and duration
            status_color = self._get_status_color(group.response_log.status_code)
            title.append(f"[{group.response_log.status_code}] ", style=status_color)
            # Format duration: show as seconds if >= 1000ms, otherwise ms
            if group.duration_ms >= 1000:
                title.append(f"{group.duration_ms / 1000:.1f}s", style="dim")
            else:
                title.append(f"{group.duration_ms:.1f}ms", style="dim")
        else:
            # Use dim white to indicate ongoing status
            title.append("ongoing", style="dim white")

        return title

    def _create_panel_content(self, group: RequestGroup) -> Group:
        """Create the panel content with route logs."""
        content_parts: List[RenderableType] = []

        # Separate background and regular logs
        regular_logs = [log for log in group.route_logs if not log.background]
        background_logs = [log for log in group.route_logs if log.background]

        # Add regular logs first
        for i, route_log in enumerate(regular_logs):
            route_line = self._format_route_line(route_log, group.start_timestamp)
            content_parts.append(route_line)

            if route_log.exc_info:
                exc_renderable = self._format_exception(route_log.exc_info)
                if exc_renderable is not None:
                    content_parts.append(Text("\n"))
                    content_parts.append(exc_renderable)

        # Add background task divider and logs if any
        if background_logs:
            content_parts.append(
                Rule(
                    "[dim blue]── Background[/dim blue]",
                    style="dim blue",
                    align="left",
                )
            )

            for i, route_log in enumerate(background_logs):
                route_line = self._format_route_line(route_log, group.start_timestamp)
                content_parts.append(route_line)

                if route_log.exc_info:
                    exc_renderable = self._format_exception(route_log.exc_info)
                    if exc_renderable is not None:
                        content_parts.append(Text("\n"))
                        content_parts.append(exc_renderable)

        # Add response-level exception if present
        if group.response_log and group.response_log.exc_info:
            exc_renderable = self._format_exception(group.response_log.exc_info)
            if exc_renderable is not None:
                # if content_parts:
                #     content_parts.append(Text("\n"))
                content_parts.append(exc_renderable)

        return Group(*content_parts)

    def _create_single_line(self, group: RequestGroup) -> Text:
        """Create a single-line format for requests without route logs."""
        line = Text()

        # Leading spaces for alignment with panel (3 spaces to match panel indentation)
        line.append("   ")

        # Determine base styling - use dim + status color for completed requests, dim for ongoing
        if group.response_log is not None:
            base_style = f"dim {self._get_status_color(group.response_log.status_code)}"
        else:
            base_style = "dim white"

        # Timestamp (dimmed with status color for completed requests)
        dt = datetime.fromtimestamp(group.start_timestamp, tz=timezone.utc)
        time_str = dt.strftime("%H:%M:%S.%f")[:-3]
        line.append(f"{time_str} | ", style=base_style)

        # Request ID (dimmed with status color for completed requests)
        line.append(f"{group.request_id[:8]} | ", style=base_style)

        if group.request_log:
            # HTTP method (padded to 7 chars)
            method_color = self.HTTP_METHOD_COLORS.get(
                group.request_log.method, "white"
            )
            line.append(f"{group.request_log.method:<7} ", style=f"bold {method_color}")

            # Path (padded to 28 characters for alignment)
            line.append(f"{group.request_log.path:<50} ", style="white")

        if group.response_log is not None:
            # Status code and duration
            status_color = self._get_status_color(group.response_log.status_code)
            line.append(f"[{group.response_log.status_code}] ", style=status_color)
            # Format duration: show as seconds if >= 1000ms, otherwise ms (dimmed with status color)
            if group.duration_ms >= 1000:
                line.append(f"{group.duration_ms / 1000:.1f}s", style=base_style)
            else:
                line.append(f"{group.duration_ms:.1f}ms", style=base_style)
        else:
            # Indicate ongoing status
            line.append("ongoing", style=base_style)

        return line

    def _format_route_line(
        self, record: RouteLogRecord, start_timestamp: float
    ) -> Text:
        """Format a route log line with relative timestamp."""
        line = Text()

        # Relative timestamp (adaptive format, right-aligned in 12 chars)
        relative_ms = (record.timestamp - start_timestamp) * 1000
        if relative_ms >= 1000:
            time_str = f"+{relative_ms / 1000:.1f}s"
        else:
            time_str = f"+{relative_ms:.0f}ms"
        line.append(f"{time_str:>13} ", style="dim")

        # Log level with buffered indicator (padded to 8 chars)
        level_color = self.LOG_LEVEL_COLORS.get(record.level, "white")
        level_text = record.level.name
        if record.buffered:
            level_text = f"*{level_text}"  # Asterisk indicates buffered logs
        line.append(f"| {level_text:<8} | ", style=level_color)

        # Message
        line.append(f"{record.message:18}", style="white")

        # Context parameters (excluding special keys)
        if record.context:
            ctx_items = [
                f"{key}={value}"
                for key, value in record.context.items()
                if key not in ("route", "handler")
            ]
            if ctx_items:
                line.append(f" [{', '.join(ctx_items)}]", style="dim")

        return line

    def _format_exception(self, exc_info: tuple) -> Optional[Traceback]:
        """Format exception traceback using Rich's pretty printing."""
        if not exc_info:
            return None

        return Traceback.from_exception(
            exc_info[0],
            exc_info[1],
            exc_info[2],
            # show_locals=True,
            # max_frames=2,
            suppress=[__file__],  # Suppress frames from this file
            width=self.live.console.width,
        )

    def _get_status_color(self, status_code: int) -> str:
        """Get color for HTTP status code."""
        if 200 <= status_code < 300:
            return self.STATUS_CODE_COLORS["2xx"]
        elif 300 <= status_code < 400:
            return self.STATUS_CODE_COLORS["3xx"]
        elif 400 <= status_code < 500:
            return self.STATUS_CODE_COLORS["4xx"]
        else:
            return self.STATUS_CODE_COLORS["5xx"]



from typing import Any, Callable, Collection, Mapping, Sequence, override

from starlette.datastructures import Headers
from starlette.exceptions import HTTPException
from starlette.middleware import Middleware
from starlette.responses import PlainTextResponse
from starlette.routing import Match, Route, get_name
from starlette.types import Receive, Scope, Send

from stario.datastar.adapters import detached_command
from stario.datastar.adapters import handler as datastar_handler
from stario.types import (
    AdapterFunction,
    HeadersConstraint,
)


class StarRoute[T](Route):

    def __init__(
        self,
        path: str,
        endpoint: Callable[..., T],
        *,
        methods: Collection[str] | None = None,
        name: str | None = None,
        include_in_schema: bool = True,
        middleware: Sequence[Middleware] | None = None,
        # Stario specific
        headers: HeadersConstraint | None = None,
        adapter: AdapterFunction[T] = datastar_handler(),
    ) -> None:

        # fmt: off
        super().__init__(
            path              = path,
            endpoint          = adapter(endpoint),
            methods           = methods,
            name              = name or get_name(endpoint),
            include_in_schema = include_in_schema,
            middleware        = middleware,
        )
        # fmt: on

        # TODO: Should I consider query=, accepts=, content_type= constraints as typed parameters?

        if headers is None:
            self.headers = {}
        elif isinstance(headers, Mapping):
            self.headers = dict(headers)
        else:
            self.headers = {}
            for h in headers:

                if isinstance(h, tuple):
                    self.headers[h[0]] = h[1]
                else:
                    self.headers[h] = None

        # self.wrapper = wrapper

    def _headers_match(self, headers: Headers) -> bool:
        for k, v in self.headers.items():

            if k not in headers:
                return False

            if v is not None and headers.get(k) != v:
                return False

        return True

    @override
    def matches(self, scope: Scope) -> tuple[Match, Scope]:
        # We override this only because we want to support headers constraint

        base_match, base_scope = super().matches(scope)

        # This would fail anyways so we can just return here
        if not self.headers or base_match != Match.FULL:
            return base_match, base_scope

        # I just hope this is light enough so we can create this over and over again
        headers = Headers(scope=scope)
        if not self._headers_match(headers):
            return Match.PARTIAL, base_scope

        # If it's a match anyways, return it :)
        return base_match, base_scope

    @override
    async def handle(self, scope: Scope, receive: Receive, send: Send) -> None:

        # If there's nothing to handle additionally, just call the super
        if not self.headers:
            await super().handle(scope, receive, send)
            return

        # If there's something to handle, we need to check the headers
        headers = Headers(scope=scope)
        if not self._headers_match(headers):

            headers_str = ", ".join(str(h) for h in self.headers)
            msg = f"Expected the following headers to be present: {headers_str}"

            if "app" in scope:
                raise HTTPException(status_code=400, detail=msg)
            else:
                response = PlainTextResponse(msg, status_code=400)
            await response(scope, receive, send)
            return

        # All good, call the super
        await super().handle(scope, receive, send)

    @override
    def __eq__(self, other: Any) -> bool:
        return super().__eq__(other) and list(self.headers) == list(other.headers)

    @override
    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        methods = sorted(self.methods or [])
        path, name = self.path, self.name
        return f"{class_name}(path={path!r}, name={name!r}, methods={methods!r}, headers={self.headers!r})"


class Query[T](StarRoute[T]):
    def __init__(
        self,
        path: str,
        endpoint: Callable[..., T],
        *,
        methods: Collection[str] | None = ["GET"],
        name: str | None = None,
        include_in_schema: bool = True,
        middleware: Sequence[Middleware] | None = None,
        # Stario specific
        headers: HeadersConstraint | None = None,
        adapter: AdapterFunction[T] = datastar_handler(),
    ) -> None:

        # fmt: off
        super().__init__(
            path              = path,
            endpoint          = endpoint,
            methods           = methods,
            name              = name,
            include_in_schema = include_in_schema,
            middleware        = middleware,
            headers           = headers,
            adapter           = adapter,
        )
        # fmt: on


class Command[T](StarRoute[T]):
    def __init__(
        self,
        path: str,
        endpoint: Callable[..., T],
        *,
        methods: Collection[str] | None = ["POST"],
        name: str | None = None,
        include_in_schema: bool = True,
        middleware: Sequence[Middleware] | None = None,
        # Stario specific
        headers: HeadersConstraint | None = None,
        adapter: AdapterFunction[T] = datastar_handler(),
    ) -> None:

        # fmt: off
        super().__init__(
            path              = path,
            endpoint          = endpoint,
            methods           = methods,
            name              = name,
            include_in_schema = include_in_schema,
            middleware        = middleware,
            headers           = headers,
            adapter           = adapter,
        )
        # fmt: on


class DetachedCommand[T](StarRoute[T]):
    def __init__(
        self,
        path: str,
        endpoint: Callable[..., T],
        *,
        methods: Collection[str] | None = ["POST"],
        name: str | None = None,
        include_in_schema: bool = True,
        middleware: Sequence[Middleware] | None = None,
        # Stario specific
        headers: HeadersConstraint | None = None,
        adapter: AdapterFunction[T] = detached_command(),
    ) -> None:

        # fmt: off
        super().__init__(
            path              = path,
            endpoint          = endpoint,
            methods           = methods,
            name              = name,
            include_in_schema = include_in_schema,
            middleware        = middleware,
            headers           = headers,
            adapter           = adapter,
        )
        # fmt: on



from stario.html import b, body, div, head, html, meta, pre, script, title
from stario.html.core import HtmlElement


def toy_inspector() -> HtmlElement:
    """
    We simply add a div positioned absolutely on the top right of the page
    with the label of key-binding opening the debug panel (CMD+P by default)
    and the content of the debug panel being a pre tag with the attribute
    data-json-signals.

    Should be slightly reduced opacity.

    <pre data-json-signals></pre>
    https://data-star.dev/reference/attributes#data-json-signals
    """

    return div(
        {
            "style": {
                "position": "absolute",
                "top": "1rem",
                "right": "1rem",
                "opacity": "0.95",
                "border": "1px solid #ccc",
                "background": "#fff",
                "padding": "0.75rem",
                "min-width": "220px",
                "z-index": "1000",
            },
        },
        b("Debug Inspector:"),
        pre(
            {
                "data-json-signals": True,
                "style": {
                    "background": "#f4f4f4",
                    "border": "1px solid #eee",
                    "padding": "0.5rem",
                    "margin-bottom": "0.25rem",
                    "font-size": "0.95em",
                    "max-height": "200px",
                    "overflow": "auto",
                },
            }
        ),
    )


load_datastar = script(
    {
        "type": "module",
        "src": "https://cdn.jsdelivr.net/gh/starfederation/datastar@main/bundles/datastar.js",
    },
)


def toy_page(
    *content: HtmlElement,
    page_title: str = "Playground",
) -> HtmlElement:

    return html(
        head(
            meta({"charset": "UTF-8"}),
            title(page_title),
            load_datastar,
        ),
        body(
            *content,
            toy_inspector(),
        ),
    )



import asyncio
from dataclasses import dataclass
from typing import AsyncGenerator, Callable, ClassVar, Generator

from starlette.background import BackgroundTask
from starlette.requests import Request
from starlette.responses import HTMLResponse, Response, StreamingResponse
from starlette.types import Receive, Scope, Send

from stario.dependencies import Dependency, DependencyLifetime
from stario.html.core import render
from stario.types import AdapterFunction, EndpointFunction, RequestHandler

from .events import patch_to_sse


@dataclass(slots=True)
class _StarioAdapter:
    """
    High-performance request adapter that converts handler functions into ASGI-compatible request handlers.

    This adapter handles dependency injection, response type detection, and content rendering
    with optimized fast paths for common response patterns.
    """

    dependencies: Dependency
    renderer: Callable[..., str]

    SSE_HEADERS: ClassVar[dict[str, str]] = {
        "Cache-Control": "no-cache",
        "Connection": "keep-alive",
        "X-Accel-Buffering": "no",
    }

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        Process a request through the dependency injection system and return an appropriate response.

        Handles multiple response types:
        - Direct Response objects (passed through unchanged)
        - Generators/AsyncGenerators (converted to SSE streams)
        - Other content (rendered to HTML)
        - None (returns 204 No Content)
        """

        # Create request for easier access
        request = Request(scope, receive, send)

        try:
            # Resolve dependencies
            content = await self.dependencies.resolve(request)

            # Fast path: direct Response return (most common case)
            if isinstance(content, Response):
                response = content

            # Fast path: Generator (SSE streaming)
            elif isinstance(content, Generator):
                response = StreamingResponse(
                    content=(patch_to_sse(item, self.renderer) for item in content),
                    media_type="text/event-stream",
                    headers=self.SSE_HEADERS,
                )

            # Fast path: AsyncGenerator (SSE streaming)
            elif isinstance(content, AsyncGenerator):
                response = StreamingResponse(
                    content=(
                        patch_to_sse(item, self.renderer) async for item in content
                    ),
                    media_type="text/event-stream",
                    headers=self.SSE_HEADERS,
                )

            # Fast path: None content
            elif content is None:
                response = HTMLResponse(content=None, status_code=204)

            # Render content and return HTML response
            else:
                rendered = self.renderer(content)
                response = HTMLResponse(content=rendered, status_code=200)

            # Send response
            await response(scope, receive, send)

        finally:

            # Cleanup is guaranteed to run here, even on exception or early return
            # Reverse order to ensure LIFO cleanup (last dependency in, first out)
            for cleanup in reversed(getattr(request.state, "cleanups", [])):
                try:
                    if cleanup[0]:
                        await cleanup[1](None, None, None)
                    else:
                        cleanup[1](None, None, None)
                except Exception:
                    # Log cleanup exceptions but don't interrupt other cleanups
                    continue


@dataclass(slots=True)
class _DetachedCommandAdapter:
    """
    Adapter for fire-and-forget operations that resolves all dependencies during request handling
    and delegates the handler execution to a background task.

    Returns 204 No Content immediately if all dependencies resolve successfully.
    Returns error response if any dependency fails.
    """

    dependencies: Dependency
    renderer: Callable[..., str]

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        Process a request by resolving all dependencies and delegating handler to background.

        Flow:
        1. Resolve all dependencies while request is being processed
        2. If successful, schedule handler as background task and return 204
        3. If dependency resolution fails, return error response (from exception)
        """

        # Create request for easier access
        request = Request(scope, receive, send)

        try:
            if len(self.dependencies.children) == 0:
                arguments = {}

            elif len(self.dependencies.children) == 1:
                # Single child - no need for TaskGroup
                child = self.dependencies.children[0]
                arguments = {child.name: await child.resolve(request)}

            else:
                # Multiple children - use gather for parallel execution
                results = await asyncio.gather(
                    *[d.resolve(request) for d in self.dependencies.children],
                    return_exceptions=True,
                )
                # Check for exceptions and raise the first one found
                for result in results:
                    if isinstance(result, Exception):
                        raise result
                arguments = {
                    c.name: result
                    for c, result in zip(self.dependencies.children, results)
                }

            # We know that no exceptions were raised, so we can create the background task
            response = HTMLResponse(
                content=None,
                status_code=204,
                background=BackgroundTask(self.dependencies.function, **arguments),
            )
            await response(scope, receive, send)

        finally:

            # Cleanup is guaranteed to run here, even on exception or early return
            # Reverse order to ensure LIFO cleanup (last dependency in, first out)
            for cleanup in reversed(getattr(request.state, "cleanups", [])):
                try:
                    if cleanup[0]:
                        await cleanup[1](None, None, None)
                    else:
                        cleanup[1](None, None, None)
                except Exception:
                    # Log cleanup exceptions but don't interrupt other cleanups
                    continue


def handler[T](
    lifetime: DependencyLifetime = "request",
    renderer: Callable[..., str] = render,
) -> AdapterFunction[T]:
    """
    This decorator factory returns decorator that is able to convert functions into request handlers.
    We apply dependency injection and response processing to the handler function.

    def foo() -> str:
        return "foo"

    decorator = handler()

    request_handler = decorator(foo)
    # request_handler is now a Starlette-compatible request foo(Request) -> Response

    """

    def decorator(handler_func: EndpointFunction[T]) -> RequestHandler:
        """
        Convert a handler function into a Starlette-compatible request handler.

        The handler function will receive resolved dependencies as arguments
        and can return various response types that will be automatically processed.
        """

        # Dependencies can be build once and just used on every request
        dependencies = Dependency.build(handler_func, lifetime)

        return _StarioAdapter(dependencies, renderer)

    return decorator


def detached_command[T](
    lifetime: DependencyLifetime = "request",
    renderer: Callable[..., str] = render,
) -> AdapterFunction[T]:
    """
    Decorator for fire-and-forget commands that resolves dependencies during request handling
    and delegates handler execution to a background task.

    Returns 204 No Content immediately if all dependencies resolve successfully.
    Returns error response if any dependency fails.

    This is useful for:
    - Long-running operations that don't need to block the response
    - Background jobs triggered by HTTP requests
    - Operations where you want to validate dependencies but not wait for execution

    Example:

        @app.post("/send-email", detached_command())
        async def send_email(user_id: PathParam[int], mailer: Annotated[Mailer, get_mailer]):
            # Dependencies are resolved before this runs
            # Handler executes in background
            # User gets 204 immediately
            await mailer.send_welcome_email(user_id)

    """

    def decorator(handler_func: EndpointFunction[T]) -> RequestHandler:
        """
        Convert a handler function into a detached-execution request handler.

        The handler function's dependencies are resolved during request processing,
        but the handler itself runs in a background task after 204 is returned.
        """

        # Dependencies can be built once and just used on every request
        dependencies = Dependency.build(handler_func, lifetime)

        return _DetachedCommandAdapter(dependencies, renderer)

    return decorator
